import json
import os
import re
import traceback
from typing import Iterator, TypeVar

import git
import jsonpatch
import yaml
from diskcache import Cache
from fastapi import Header, HTTPException
from github import Github
from github.GithubException import GithubException
from loguru import logger

from sweepai.config.client import SweepConfig
from sweepai.config.server import (
    CACHE_DIRECTORY,
    CURL_CA_BUNDLE,
    GITHUB_API_BASE_URL,
    GITHUB_BASE_URL,
    WHITELISTED_USERS,
)
from sweepai.core.entities import FileChangeRequest, Snippet
from sweepai.core.github_utils import (
    ClonedRepo,
    CustomGithub,
    MockClonedRepo,
    get_github_client,
    get_installation_id,
)
from sweepai.dataclasses.chat_agent_context import ChatAgentContext
from sweepai.dataclasses.file_modification_state import FileModificationState
from sweepai.modify.modify_utils import get_latest_contents
from sweepai.o11y.posthog_trace import posthog_trace
from sweepai.review.review_utils import split_diff_into_patches  # 0.4s import
from sweepai.search.agent.agent_utils import search_codebase
from sweepai.search.agent.search_prompts import pr_format
from sweepai.search.agent.summarize_directory import (
    recursively_summarize_directory,
    summarize_frameworks,
    summary_caches,
)
from sweepai.utils.diff import generate_diff, remove_whitespace_changes
from sweepai.utils.str_utils import pack_items_for_prompt, strip_triple_quotes
from sweepai.web.webhook_secret import hash_sha256

auth_cache = Cache(f"{CACHE_DIRECTORY}/auth_cache")
repo_cache = f"{CACHE_DIRECTORY}/repos"
message_cache = f"{CACHE_DIRECTORY}/messages"

T = TypeVar("T")

def get_cached_installation_id(org_name: str) -> str:
    return get_installation_id(org_name)


def get_github_client_from_org(org_name: str) -> tuple[str, CustomGithub]:
    return get_github_client(get_cached_installation_id(org_name))


def get_authenticated_github_client(repo_name: str, access_token: str):
    # Returns read access, write access, or none
    if access_token == "null":
        raise Exception("Access token is null")
    g = Github(access_token, base_url=GITHUB_API_BASE_URL)
    user = g.get_user()
    try:
        repo = g.get_repo(repo_name)
        return g
    except Exception:
        org_name, _ = repo_name.split("/")
        try:
            _token, g = get_github_client_from_org(org_name)
        except Exception:
            raise HTTPException(
                status_code=404,
                detail=f"Could not fetch installation for {repo_name}. Make sure to install the app on this repository."
            )
        try:
            repo = g.get_repo(repo_name)
        except Exception as e:
            if "Not Found" in str(e):
                raise HTTPException(
                    status_code=404,
                    detail=f"Repository {repo_name} not found or not installed. Make sure to install the app on this repository."
                )
            raise HTTPException(
                status_code=400,
                detail=f"Error getting repo {repo_name}: {e}"
            )
        if repo.has_in_collaborators(user.login) or user.login in WHITELISTED_USERS:
            return g
        else:
            raise HTTPException(
                status_code=403,
                detail=f"User {user.login} does not have the necessary permissions for the repository {repo_name}."
            )


async def get_token_header(authorization: str = Header(...)):
    if not authorization.startswith("Bearer "):
        raise HTTPException(status_code=400, detail="Invalid token")
    return authorization.removeprefix("Bearer ")


def strip_branch_prefix(branch_name: str):
    return (
        branch_name.removeprefix("refs/heads/")
        .removeprefix("refs/tags/")
        .removeprefix("refs/remotes/")
        .removeprefix("origin/")
        .removeprefix("refs/")
    )


def check_valid_commit_sha(cloned_repo: ClonedRepo, commit_sha: str):
    try:
        cloned_repo.git_repo.git.cat_file(e=commit_sha)
        return True
    except git.GitCommandError:
        return False


def get_branch(cloned_repo: ClonedRepo, branch: str):
    try:
        return cloned_repo.github_repo.get_branch(branch)
    except GithubException as e:
        if e.status == 404:
            return None
        raise e


@posthog_trace
def check_repo_exists(
    username: str,
    repo_name: str,
    access_token: str,
    metadata: dict = {},
    do_clone: bool = False,
):
    # Can be refactored and re-used logic
    org_name, repo = repo_name.split("/")
    if os.path.exists(f"{repo_cache}/{repo}"):
        return {"success": True}
    try:
        if do_clone:
            print(f"Cloning {repo_name} to {repo_cache}/{repo}")
        else:
            print(f"Checking {repo_name}")
        os.environ["GIT_CURL_VERBOSE"] = "1"
        if do_clone or CURL_CA_BUNDLE:  # always clone with CURL_CA_BUNDLE, ls-remote does not work with it
            print("Cloning...")
            ClonedRepo.from_repo_full_name(repo_full_name=repo_name)
        if do_clone:
            print(f"Cloned {repo_name} to {repo_cache}/{repo}")
        else:
            print(f"Checked {repo_name}")
        return {"success": True}
    except git.GitCommandError as e:
        error_message = str(e)
        if "Repository not found" in error_message:
            return {"success": False, "error": "Repository not found"}
        elif "Invalid username or password" in error_message:
            return {"success": False, "error": "Invalid permissions"}
        else:
            return {"success": False, "error": f"Git error: {error_message}"}
    except Exception as e:
        return {"success": False, "error": str(e)}


def size_of_largest_change(patch: str):
    patches = split_diff_into_patches(patch, "")
    num_changes_per_patch = [patch.changes.count("\n+") + patch.changes.count("\n-") for patch in patches]
    return max(num_changes_per_patch) if num_changes_per_patch else 0


def get_pr_snippets(
    repo_name: str,
    annotations: dict,
    cloned_repo: ClonedRepo,
    context: ChatAgentContext = ChatAgentContext(use_case="ticket"),
):
    pr_snippets_with_change_size = []
    skipped_pr_snippets = []
    sweep_config = SweepConfig()
    pulls = annotations.get("pulls", [])
    pulls_messages: str = ""
    renamed_files: list[str] = []
    for pull in pulls:
        patch = pull["file_diffs"]
        diff_patch = ""
        # Filters copied from get_pr_changes
        for file_data in patch:
            file_path = file_data["filename"]
            if sweep_config.is_file_excluded(file_path):
                continue
            try:
                file_contents = cloned_repo.get_file_contents(file_path)
            except FileNotFoundError:
                logger.warning(f"Error getting file contents for {file_path}")
                continue
            is_file_suitable, reason = sweep_config.is_file_suitable(file_contents)
            if not is_file_suitable:
                continue
            diff = file_data.get("patch", "")
            if file_data["status"] == "renamed":
                new_file_name = file_data["filename"]
                old_file_name = file_data["previous_filename"]
                renamed_files.append(f"The file {old_file_name} was renamed to {new_file_name}.")
            elif file_data["status"] == "added":
                pr_snippets_with_change_size.append(
                    (
                        len(file_contents.split("\n")),
                        Snippet.from_file(file_path, file_contents, score=1),
                    )
                )
            elif file_data["status"] == "modified":
                patches = split_diff_into_patches(diff, file_path)
                num_changes_per_patch = [patch.changes.count("\n+") + patch.changes.count("\n-") for patch in patches]
                # Either the largest change is > 25 lines, or it changes more than 10% of the file
                if (
                    max(num_changes_per_patch, default=0) > 25
                    or file_data["changes"] / (file_contents.count("\n") + 1) > 0.1
                ):
                    # print(f"adding {file_path}")
                    # print(num_changes_per_patch)
                    # print(file_contents.count("\n"))
                    pr_snippets_with_change_size.append(
                        (
                            max(num_changes_per_patch, default=0),
                            Snippet.from_file(file_path, file_contents, score=1),
                        )
                    )
                else:
                    skipped_pr_snippets.append(Snippet.from_file(file_path, file_contents))
            if file_data["status"] in ("added", "modified", "removed"):
                diff_patch += f"File: {file_path}\n" + diff.strip("\n") + "\n\n"
        if diff_patch:
            pulls_messages += (
                pr_format.render(
                    title=pull["title"],
                    body=pull["body"],
                    patch=remove_whitespace_changes(diff_patch).strip("\n"),
                    url=f"https://{GITHUB_BASE_URL}/{repo_name}/pull/{pull['number']}",
                    use_case=context.use_case,
                )
                + "\n\n"
            )
            if renamed_files:
                renamed_files_string = "\n".join(renamed_files)
                pulls_messages += f"<renamed_files>{renamed_files_string}</renamed_files>\n\n"
    pr_snippets_with_change_size = sorted(pr_snippets_with_change_size, key=lambda x: x[0], reverse=True)
    pr_snippets = [snippet for _, snippet in pr_snippets_with_change_size]
    pr_snippets = pack_items_for_prompt(pr_snippets, lambda x: x.render(i=1), 100_000)
    return pr_snippets, skipped_pr_snippets, pulls_messages


# this is messy; its modular so we can move it elsewhere later
def get_repo_specific_description(cloned_repo: MockClonedRepo):
    try:
        sweep_yaml_contents = cloned_repo.get_file_contents("sweep.yaml")
        sweep_yaml = yaml.safe_load(sweep_yaml_contents)
        description = sweep_yaml.get("description", "")
        system_prompt_formatted_description = f"\nThis is a user provided description of the codebase. Keep this in mind if it is relevant to their query:\n<codebase_description>\n{description}\n</codebase_description>\n"
        return system_prompt_formatted_description
    except FileNotFoundError:
        logger.info(f"No .sweep.yaml file present in {cloned_repo.repo_full_name}.")
        return ""
    except Exception as e:
        logger.error(f"Error reading .sweep.yaml file: {e}")
        return ""


def stream_state_diff(
    stream: Iterator[T],
    initial_state: T = [],
) -> Iterator[T]:
    previous_state = initial_state
    try:
        for current_state in stream:
            patch = jsonpatch.JsonPatch.from_diff(previous_state, current_state)
            if patch:
                yield patch.to_string()
            previous_state = current_state
    except (Exception, KeyboardInterrupt) as e:
        yield json.dumps(
            [
                {
                    "status": "error",
                    "error": f"{str(e)}\n{str(traceback.format_exc())}",
                    "traceback": str(traceback.format_exc()),
                }
            ]
        )
        raise


def clean_code_suggestions(original_code: str, new_code: str, file_path: str, cloned_repo: ClonedRepo):
    try:
        file_contents = cloned_repo.get_file_contents(file_path)
    except FileNotFoundError:
        return original_code, new_code
    # this handles the case where the original code is a comment like "// the rest of the tests above"
    if (
        original_code.strip().count("\n") == 0
        and original_code not in file_contents
        and (original_code.strip().startswith("#") or original_code.strip().startswith("//"))
        and new_code.count("\n") > 0
    ):
        if original_code in new_code:
            return "", new_code.replace(original_code + "\n", "")
        return "", new_code
    return original_code, new_code


def extract_and_clean_code_suggestions(raw_suggestions: str, cloned_repo: ClonedRepo):
    pattern = r"<code_change>\s*<file_path>\n*(?P<filePath>[\s\S]+?)\n*</file_path>\s*(<justification>\n*(?P<justification>[\s\S]+?)\n*</justification>)?\s*(<original_code>\n*(?P<originalCode>[\s\S]*?)</original_code>\n*)?<new_code>\n*(?P<newCode>[\s\S]+?)\n*(</new_code>)?\s*</code_change>"
    # combine additions of the same file together
    code_suggestions_raw = [
        {
            "filePath": match_.groupdict().get("filePath", ""),
            "originalCode": strip_triple_quotes(match_.groupdict().get("originalCode") or ""),
            "newCode": strip_triple_quotes(match_.groupdict().get("newCode") or ""),
        }
        for match_ in re.finditer(pattern, raw_suggestions, re.DOTALL)
    ]
    new_code_suggestions_raw = []
    for code_suggestion in code_suggestions_raw:
        fcr = next(
            (
                fcr
                for fcr in new_code_suggestions_raw
                if fcr["filePath"] == code_suggestion["filePath"]
                and fcr["originalCode"] == code_suggestion["originalCode"] == ""
            ),
            None,
        )
        file_exists = False
        try:
            cloned_repo.get_file_contents(code_suggestion["filePath"])
            file_exists = True
        except FileNotFoundError:
            file_exists = False
        if fcr and not file_exists:
            fcr["newCode"] += "\n\n" + strip_triple_quotes(code_suggestion["newCode"])
        else:
            new_code_suggestions_raw.append(code_suggestion)
    code_suggestions_raw = new_code_suggestions_raw
    if code_suggestions_raw:
        code_suggestions = []
        for code_suggestion in code_suggestions_raw:
            original_code, new_code = clean_code_suggestions(
                strip_triple_quotes(code_suggestion["originalCode"]),
                strip_triple_quotes(code_suggestion["newCode"]),
                code_suggestion["filePath"],
                cloned_repo,
            )
            code_suggestions.append(
                {
                    "filePath": code_suggestion["filePath"],
                    "originalCode": original_code,
                    "newCode": new_code,
                    "state": "pending",
                    "error": None,
                    "id": "id_" + hash_sha256(code_suggestion["filePath"] + original_code + new_code),
                }
            )
        return code_suggestions
    return []


def validate_code_suggestions(
    code_suggestions: list,
    cloned_repo: ClonedRepo,
    modify_files_dict: FileModificationState,
    current_content: str,
):
    file_change_requests = []
    for code_suggestion in code_suggestions:
        try:
            file_contents = get_latest_contents(code_suggestion["filePath"], cloned_repo, modify_files_dict)
            if not file_contents:
                raise FileNotFoundError(f"File {code_suggestion['filePath']} not found.")
            change_type = "modify"
        except FileNotFoundError:
            change_type = "create"
        file_change_requests.append(
            FileChangeRequest(
                filename=code_suggestion["filePath"],
                instructions=f"<original_code>\n{code_suggestion['originalCode']}\n</original_code>\n<new_code>\n{code_suggestion['newCode']}\n</new_code>",
                change_type=change_type,
            )
        )
    return file_change_requests, current_content


def render_code_suggestions_for_logging(code_suggestions: list):
    """Render the diffs from old_code and new_code with file_path above"""
    rendered_suggestions = ""
    for suggestion in code_suggestions:
        suggestion_diff = generate_diff(
            suggestion["originalCode"], suggestion["newCode"]
        )
        suggestion_diff = suggestion_diff if suggestion_diff else "NO DIFF GENERATED"
        rendered_suggestion = f"File: {suggestion['filePath']}\n" + suggestion_diff
        rendered_suggestions += rendered_suggestion + "\n\n"
    return rendered_suggestions.strip()


def index_repo(repo_name: str, access_token: str):
    cloned_repo = ClonedRepo(repo_name, installation_id=-1, token=access_token)
    search_codebase("a", cloned_repo)
    recursively_summarize_directory([], cloned_repo)
    summarize_frameworks([], cloned_repo)


def get_directory_summaries(repo_name: str, directories: list[str]) -> dict[str, str]:
    key = ("summaries", repo_name)
    if key not in summary_caches:
        return {directory: "" for directory in directories}

    summaries, _ = summary_caches[key]
    return {directory: summaries.get(directory, "") for directory in directories}


if __name__ == "__main__":
    index_repo(os.environ.get("REPO"), os.environ.get("GITHUB_PAT"))



import fnmatch
import multiprocessing
import os
import re
import subprocess
import time
from collections.abc import Iterable

from diskcache import Cache
from loguru import logger
from tantivy import Document, Index, SchemaBuilder
from tqdm import tqdm

from sweepai.config.client import SweepConfig
from sweepai.config.server import CACHE_DIRECTORY
from sweepai.core.entities import Snippet
from sweepai.dataclasses.files import LexicalSearchDocument
from sweepai.o11y.event_logger import posthog
from sweepai.search.agent.files_summarizer import file_summary_cache
from sweepai.search.query.repo_parsing_utils import directory_to_chunks
from sweepai.search.query.vector_db import multi_get_query_texts_similarity
from sweepai.utils.cache import hash_code
from sweepai.utils.streamable_functions import streamable
from sweepai.utils.timer import Timer

token_cache = Cache(f"{CACHE_DIRECTORY}/token_cache")  # we instantiate a singleton, diskcache will handle concurrency
lexical_index_cache = Cache(f"{CACHE_DIRECTORY}/lexical_index_cache")
snippets_cache = Cache(f"{CACHE_DIRECTORY}/snippets_cache")
CACHE_VERSION = "v1.0.17"

# pylint: disable=no-member
schema_builder = SchemaBuilder()
schema_builder.add_text_field("title", stored=True)
schema_builder.add_text_field("body", stored=True)
schema_builder.add_integer_field("doc_id", stored=True)
schema = schema_builder.build()
# pylint: enable=no-member


class CustomIndex:
    def __init__(self, cache_path: str = None):
        os.makedirs(cache_path, exist_ok=True)
        self.index = Index(schema)  # pylint: disable=no-member

    def add_documents(self, documents: Iterable):
        writer = self.index.writer()
        for doc_id, (title, text) in enumerate(documents):
            writer.add_document(Document(title=title, body=text, doc_id=doc_id))  # pylint: disable=no-member
        writer.commit()

    def search_index(self, query: str) -> list[tuple[str, float, dict]]:
        tokenized_query = tokenize_code(query)
        parsed_query = self.index.parse_query(tokenized_query)
        searcher = self.index.searcher()  # for some reason, the first searcher is empty
        for i in range(100):
            searcher = self.index.searcher()
            if searcher.num_docs > 0:
                break
            print(f"Index is empty, sleeping for {0.01 * i} seconds")
            time.sleep(0.01)
        else:
            raise Exception("Index is empty")
        results = searcher.search(parsed_query, limit=200).hits
        return [(searcher.doc(doc_id)["title"][0], score, searcher.doc(doc_id)) for score, doc_id in results]


variable_pattern = re.compile(r"([A-Z][a-z]+|[a-z]+|[A-Z]+(?=[A-Z]|$))")


def tokenize_code(code: str) -> str:
    matches = re.finditer(r"\b\w{2,}\b", code)
    tokens: list[str] = []
    for m in matches:
        text = m.group()
        underscore_matches = text.split("_")
        if len(underscore_matches) > 1:
            tokens.append(text)
        for underscore_match in underscore_matches:
            variable_matches = variable_pattern.findall(underscore_match)
            # if len(variable_matches) > 1:
            #     tokens.append(underscore_match)
            for part in variable_matches:
                if len(part) < 2:
                    continue
                # if more than half of the characters are letters
                # and the ratio of unique characters to the number of characters is less than 5
                if (
                    sum(1 for c in part if "a" <= c <= "z" or "A" <= c <= "Z" or "0" <= c <= "9") > len(part) // 2
                    and len(part) / len(set(part)) < 4
                ):
                    tokens.append(part)
    return " ".join(token.lower() for token in tokens)


def snippets_to_docs(snippets: list[Snippet], len_repo_cache_dir):
    docs = []
    for snippet in snippets:
        docs.append(
            LexicalSearchDocument(
                title=f"{snippet.file_path[len_repo_cache_dir:]}:{snippet.start}-{snippet.end}",
                content=snippet.get_snippet(add_ellipsis=False, add_lines=False),
            )
        )
    return docs


@streamable
def prepare_index_from_snippets(
    snippets: list[Snippet],
    len_repo_cache_dir: int = 0,
    do_not_use_file_cache: bool = False,
    cache_path: str = None,
) -> CustomIndex | None:
    all_docs: list[LexicalSearchDocument] = snippets_to_docs(snippets, len_repo_cache_dir)
    if len(all_docs) == 0:
        posthog.capture(
            "preparing index failed with no docs",
            "no docs found for index",
            properties={
                "cache_path": cache_path,
                "do_not_use_file_cache": do_not_use_file_cache,
                "len_repo_cache_dir": len_repo_cache_dir,
                "len_snippets": len(snippets),
            },
        )
        return None
    index = CustomIndex(cache_path=cache_path)
    yield "Tokenizing documents...", index
    all_tokens = []
    try:
        with Timer() as timer:
            for doc in all_docs:
                all_tokens.append(token_cache.get(doc.content + CACHE_VERSION))
            misses = [i for i, token in enumerate(all_tokens) if token is None]
            num_workers = multiprocessing.cpu_count() // 3
            if misses:
                if num_workers > 1 or len(misses) > 1000:
                    with multiprocessing.Pool(processes=num_workers) as p:
                        missed_tokens = p.map(
                            tokenize_code,
                            tqdm(
                                [all_docs[i].content for i in misses],
                                total=len(misses),
                                desc="Tokenizing documents",
                            ),
                        )
                else:
                    missed_tokens = [tokenize_code(all_docs[i].content) for i in misses]
                for i, token in enumerate(missed_tokens):
                    all_tokens[misses[i]] = token
                    token_cache[all_docs[misses[i]].content + CACHE_VERSION] = token
        logger.debug(f"Tokenizing documents took {timer.time_elapsed} seconds")
        yield "Building lexical index...", index
        all_titles = [doc.title for doc in all_docs]
        with Timer() as timer:
            index.add_documents(tqdm(zip(all_titles, all_tokens), total=len(all_docs), desc="Indexing"))
        logger.debug(f"Indexing took {timer.time_elapsed} seconds")
    except FileNotFoundError as e:
        logger.exception(e)
    yield "Index built", index
    return index


def search_index(query: str, index: CustomIndex):
    """Search the index based on a query.

    This function takes a query and an index as input and returns a dictionary of document IDs
    and their corresponding scores.
    """
    # Create a query parser for the "content" field of the index
    if index:
        results_with_metadata = index.search_index(query)
    else:
        results_with_metadata = []
    # Search the index
    res = {}
    for doc_id, score, _ in results_with_metadata:
        if doc_id not in res:
            res[doc_id] = score
    # min max normalize scores from 0.5 to 1
    if len(res) == 0:
        max_score = 1
        min_score = 0
    else:
        max_score = max(res.values())
        min_score = min(res.values()) if min(res.values()) < max_score else 0
    res = {k: (v - min_score) / (max_score - min_score) for k, v in res.items()}
    return res


SNIPPET_FORMAT = """<file_path>
{file_path}
</file_path>

<file_summary>
{file_summary}
</file_summary>

{contents}"""


# @file_cache(ignore_params=["snippets"])
@streamable
def compute_vector_search_scores(queries: list[str], snippets: list[Snippet]):
    # get get dict of snippet to score
    def get_file_summary_from_cache(snippet: Snippet):
        return file_summary_cache.get(hash_code(snippet.content))

    with Timer() as timer:
        snippet_str_to_contents = {
            snippet.denotation: SNIPPET_FORMAT.format(
                file_path=snippet.file_path,
                file_summary=get_file_summary_from_cache(snippet),
                contents=snippet.get_snippet(add_ellipsis=False, add_lines=False),
            )
            for snippet in snippets
        }
    logger.info(f"Snippet to contents took {timer.time_elapsed:.2f} seconds")
    snippet_contents_array = list(snippet_str_to_contents.values())
    multi_query_snippet_similarities = []
    for (
        message,
        multi_query_snippet_similarities,
    ) in multi_get_query_texts_similarity.stream(queries, snippet_contents_array):
        yield message, multi_query_snippet_similarities
    snippet_denotations = [snippet.denotation for snippet in snippets]
    snippet_denotation_to_scores = [
        {snippet_denotations[i]: score for i, score in enumerate(query_snippet_similarities)}
        for query_snippet_similarities in multi_query_snippet_similarities
    ]
    yield "Vector search scores computed.", snippet_denotation_to_scores
    return snippet_denotation_to_scores


def get_lexical_cache_key(repo_directory: str, commit_hash: str | None = None, seed: str = ""):
    commit_hash = (
        commit_hash
        or subprocess.run(
            ["git", "rev-parse", "HEAD"],
            cwd=repo_directory,
            capture_output=True,
            text=True,
        ).stdout.strip()
    )
    repo_directory = os.path.basename(repo_directory)
    return f"{repo_directory}_{commit_hash}_{CACHE_VERSION}_{seed}"


@streamable
def prepare_lexical_search_index(
    repo_directory: str,
    sweep_config: SweepConfig,
    do_not_use_file_cache: bool = False,  # choose to not cache results
    seed: str = "",  # used for lexical cache key
    globs: list[str] = ["**/*"],
):
    lexical_cache_key = get_lexical_cache_key(repo_directory, seed=seed)

    yield "Collecting snippets...", [], None
    snippets_results = snippets_cache.get(lexical_cache_key)
    if snippets_results is None:
        snippets, file_list = directory_to_chunks(
            repo_directory, sweep_config, do_not_use_file_cache=do_not_use_file_cache
        )
        snippets_cache[lexical_cache_key] = snippets, file_list
    else:
        snippets, file_list = snippets_results

    if len(snippets) == 0:
        logger.warning("No snippets found for index, repository may be empty")
        yield "No snippets found for index, repository may be empty", [], None
        return [], None

    globs = [glob.replace("/**", "").replace("*/*", "*") for glob in globs]  # /** isn't handled properly by fnmatch
    # snippets get kept if they match any of the globs
    filtered_snippets = [
        snippet
        for snippet in snippets
        if any(fnmatch.fnmatch(name=snippet.file_path[len(repo_directory) :].lstrip("/"), pat=glob) for glob in globs)
    ]

    if len(filtered_snippets) == 0:
        logger.warning("No snippets found for index, probably due to globs being too restrictive")
        yield "No snippets found for index, globs may be too restrictive", [], None
        return [], None

    snippets = filtered_snippets
    yield "Building index...", snippets, None

    index = prepare_index_from_snippets(
        snippets,
        len_repo_cache_dir=len(repo_directory) + 1,
        do_not_use_file_cache=do_not_use_file_cache,
        cache_path=f"{CACHE_DIRECTORY}/lexical_index_cache/{lexical_cache_key}",
    )

    yield "Lexical index built.", snippets, index

    return snippets, index


if __name__ == "__main__":
    repo_directory = os.getenv("REPO_DIRECTORY")
    sweep_config = SweepConfig()
    assert repo_directory
    import time

    start = time.time()
    snippets, index = prepare_lexical_search_index(repo_directory, sweep_config, None)
    breakpoint()  # noqa: T100
    result = search_index("logger export", index)
    print("Time taken:", time.time() - start)
    # print some of the keys
    print(list(result.keys())[:5])
    # print the first 2 result keys sorting by value
    print(sorted(result.items(), key=lambda x: result.get(x, 0), reverse=True)[:5])



from sweepai.core.llm.chat import LATEST_CLAUDE_MODEL, Thread, continuous_llm_calls
from sweepai.utils.str_utils import safe_extract_xml_tag

system_message = """You are a helpful assistant that summarizes the history of changes you previously made to a GitHub repository in order to fix a GitHub Actions failure."""

user_message_prompt = """<history>
{history}
</history>

<current_failing_error_logs>
{current_failing_error_logs}
</current_failing_error_logs>

For each of the {n} fix attempts, respond in the following format:

<fix_attempts>
<fix_attempt index="i">
<error>
[what was failing, copying relevant error logs]
</error>
<analysis>
[what you were thinking at this point in time]
</analysis>
<fixes>
[what did you change to address the error? highlight relevant code changes that addresses the error]
</fixes>
<conclusion>
[were you successful? otherwise, what did you incorrectly assume and what did you learn from this fix attempt?]
</conclusion>
</fix_attempt>
</fix_attempts>

# Guidelines:
- Be objective and focus on the codebase. Do not make any interpretations about what may be the root cause of the failure.
- Summarize learnings as "A does not do B", not "I incorrectly assumed A does B".
- Do not say "I didn't analyze the implementation thoroughly" or "I need to look at this closer".

Then summarize everything in the following format, writing in first-person:

<summary>
<successful_fixes>
1. [List all the fixes you made that were successful, what it accomplished and why]
2. [Repeat for each fix attempt]
</successful_fixes>
<unsuccessful_fixes>
1. [List all the fix attempts you made that were unsuccessful, the assumptions you made that were incorrect, and what you learned about the codebase. Be specific about what you tried. Focus on the codebase, do not say "I didn't analyze the implementation thoroughly".]
2. [For each unsuccessful assumption, indicate the number of the fix attempt it is associated with.]
</unsuccessful_fixes>
</summary>"""


def summarize_history(username: str, history: str, current_failing_error_logs: str) -> str:
    if not history:
        return ""
    thread = Thread.from_system_message_string(system_message)
    user_message = user_message_prompt.format(
        history=history,
        current_failing_error_logs=current_failing_error_logs,
        n=history.count("<fix_attempt"),
    )
    response = continuous_llm_calls(
        username=username,
        thread=thread,
        content=user_message,
        stop_sequences=["</summary>"],
        model=LATEST_CLAUDE_MODEL,
    )
    return safe_extract_xml_tag(response, "summary")



from datetime import datetime, timedelta, timezone
from github import Github
from loguru import logger
from tqdm import tqdm
from sweepai.config.server import GITHUB_API_BASE_URL
from sweepai.core.github_utils import github_integration


def get_merged_pr_urls(username: str, token: str | None = None, repos: list[str] | None = None) -> list[str]:
    github_client = github_integration.get_user_installation(username).get_github_for_installation() if not token else Github(token, base_url=GITHUB_API_BASE_URL)
    github_client.per_page = 100
    user = github_client.get_user(username)

    start_date = datetime.now(timezone.utc) - timedelta(days=90)

    logger.info(f"Fetching repos for {username}...")
    github_repos = []
    for repo in tqdm(user.get_repos(type="all", sort="pushed", direction="desc"), desc="Fetching repos"):
        if max(repo.pushed_at, repo.updated_at, repo.created_at) < start_date:
            break
        github_repos.append(repo)

    visited_repos = set([repo.full_name for repo in github_repos])

    events = list(tqdm(user.get_events(), desc="Fetching events"))
    for event in events:
        repo_name = event.raw_data["repo"]["name"]
        if event.created_at < start_date:
            break
        if repo_name not in visited_repos:
            github_repos.append(github_client.get_repo(repo_name))
            visited_repos.add(repo_name)
    
    for repo_name in repos:
        if repo_name and repo_name not in visited_repos:
            github_repos.append(github_client.get_repo(repo_name))
            visited_repos.add(repo_name)

    github_repos.sort(key=lambda x: max(x.pushed_at, x.updated_at, x.created_at), reverse=True)
    logger.info(f"Found {len(github_repos)} repos for {username}...")

    prs = []
    for i, repo in enumerate(github_repos):
        if max(repo.pushed_at, repo.updated_at, repo.created_at) < start_date:
            break
        logger.info(f"Fetching PRs for {repo.name} by {username}...")
        for pr in tqdm(repo.get_pulls(state="closed", sort="updated", direction="desc"), desc=f"Fetching PRs for {repo.name} ({i}/{len(github_repos)})"):
            if (pr.updated_at or pr.created_at) < start_date:
                break
            if pr.user.login == username and pr.merged_at: # can check commits as well but it's super slow
                prs.append(pr)
    urls = [pr.html_url for pr in prs]
    return urls



from __future__ import annotations

from typing import Callable

from jinja2 import BaseLoader, ChainableUndefined, Environment, StrictUndefined

from sweepai.config.server import DEV
from sweepai.core.entities import Snippet


class EnvironmentWithDecorator(Environment):
    def add_filter(self, filter_func):
        self.filters[filter_func.__name__] = filter_func
        return filter_func


env = EnvironmentWithDecorator(loader=BaseLoader(), undefined=StrictUndefined if DEV else ChainableUndefined)

env.globals.update(
    {
        "enumerate": enumerate,
    }
)

env.filters.update(
    {
        "enumerate": enumerate,
        "list": lambda values: "\n".join([f"- {value}" for value in values]),
        "wrap_xml": lambda value, tag: f"<{tag}>\n{value}\n</{tag}>",
        "wrap_xml_if_non_empty": lambda value, tag: value and f"<{tag}>\n{value}\n</{tag}>" or "",
    }
)


@env.add_filter
def render_snippets(snippets: list[Snippet], EXPAND_SIZE: int = 100):
    return "\n".join([snippet.render(i, EXPAND_SIZE=EXPAND_SIZE) for i, snippet in enumerate(snippets)])


def ordered_dedup(lst: list[str]):
    result = []
    seen = set()
    for item in lst:
        if item not in seen:
            seen.add(item)
            result.append(item)
    return result


class Prompt:
    def __init__(
        self,
        template_string: str,
        strip: bool = True,
        globals: list[Callable] = [],
        filters: list[Callable] = [],
        constants: dict[str, str] = {},
    ):
        for global_func in globals:
            env.globals[global_func.__name__] = global_func
        for filter_func in filters:
            env.filters[filter_func.__name__] = filter_func
        template_string = template_string.strip("\n") if strip else template_string
        self.template = env.from_string(template_string)
        self.constants = constants

    def render(self, strip: bool = True, **kwargs) -> str:  # written this way for backwards compatibility with format
        # ensure uniqueness and completeness of kwargs
        result = self.template.render({**self.constants, **kwargs})
        return result.strip("\n") if strip else result



from __future__ import annotations

from dataclasses import dataclass, field

from sweepai.utils.diff import generate_diff


@dataclass
class FileContents:
    original_contents: str
    new_contents: str

    @property
    def has_changes(self) -> bool:
        return self.original_contents != self.new_contents
    
    def diff(self) -> str:
        return generate_diff(self.original_contents, self.new_contents)
    
    def __repr__(self) -> str:
        return self.diff()


@dataclass
class FileModificationState:
    files: dict[str, FileContents] = field(default_factory=dict)

    def get_original_contents(self, file_path: str) -> str:
        return self.files[file_path].original_contents

    def get_new_contents(self, file_path: str) -> str:
        return self.files[file_path].new_contents

    def add_or_update_file(self, file_path: str, original_contents: str, new_contents: str):
        self.files[file_path] = FileContents(original_contents, new_contents)

    def update_original_contents(self, file_path: str, new_contents: str):
        self.files[file_path].original_contents = new_contents

    def update_new_contents(self, file_path: str, new_contents: str):
        self.files[file_path].new_contents = new_contents

    def file_has_changes(self, file_path: str) -> bool:
        return self.files[file_path].has_changes

    def dict(self) -> dict:
        """
        Dumps this object to a dictionary.
        """
        return {
            "files": {
                file_path: {
                    "original_contents": contents.original_contents,
                    "new_contents": contents.new_contents,
                }
                for file_path, contents in self.files.items()
            }
        }
    
    def to_raw_code_suggestions(self) -> list[dict[str, str]]:
        return [
            {
                "filePath": file_path,
                "originalCode": contents.original_contents,
                "newCode": contents.new_contents,
                "state": "done",
            }
            for file_path, contents in self.files.items()
        ]

    
    def __getitem__(self, file_path: str) -> FileContents:
        return self.files[file_path]
    
    def __len__(self):
        return len(self.files)

    def __contains__(self, file_path: str) -> bool:
        return file_path in self.files
    
    def __iter__(self):
        return iter(self.files)

    @classmethod
    def from_dict(cls, dict: dict) -> FileModificationState:
        # Handles legacy dicts
        file_modification_state = FileModificationState()
        for file_path, file_content in dict.items():
            file_modification_state.add_or_update_file(
                file_path, file_content["original_contents"], file_content["contents"]
            )
        return file_modification_state
    
    def diff(self) -> str:
        return "\n\n".join([f"#### {file_path}\n\n{file_contents.diff()}" for file_path, file_contents in self.files.items()])
    
    def __repr__(self) -> str:
        return self.diff()



import os
import sys
from typing import Dict, List

import joblib
from diskcache import Cache
from loguru import logger
from tqdm import tqdm

from sweepai.config.client import SweepConfig
from sweepai.config.server import CACHE_DIRECTORY, DEV
from sweepai.core.github_utils import ClonedRepo
from sweepai.core.llm.chat import call_llm
from sweepai.search.agent.summarize_directory import recursively_summarize_directory
from sweepai.search.query.repo_parsing_utils import is_actual_code_file
from sweepai.utils.cache import hash_code
from sweepai.utils.str_utils import extract_xml_tag

file_summary_cache = Cache(f"{CACHE_DIRECTORY}/file_summary_cache")

instructions = "Start by explaining the the file in the context of its directories. Focus on nuances and patterns. Then, explain what types of business logic this file contains. For each code entity, explain its purpose and what it handles. Be concise and optimize for informational density. One paragraph."

system_prompt = "Your job is to summarize the following file. " + instructions

user_prompt = (
    """First, review the parent directory summaries for context:

<directory_summaries>
{DIRECTORY_SUMMARIES}
</directory_summaries>

Now, examine the title and contents of the file:

<file_name>
{FILE_NAME}
</file_name>

<file_contents>
{FILE_CONTENTS}
</file_contents>

Provide your summary within <file_summary> tags.

<file_summary>
[Your concise, one-paragraph summary goes here]
</file_summary>

"""
    + instructions
)


class FilesSummarizer:
    def __init__(self, cloned_repo: ClonedRepo):
        self.cloned_repo = cloned_repo
        logger.info("Summarizing directories...")
        self.directory_summaries, _ = recursively_summarize_directory(self.cloned_repo)
        logger.info("Summarizing files...")
        self.file_summaries = self.generate_file_summaries()

    def get_ancestors(self, relative_path: str) -> List[str]:
        """Given a relative_path, compute all possible ancestors"""
        parts = relative_path.split(os.sep)
        ancestors = []
        for i in range(1, len(parts)):
            ancestor = os.path.join(*parts[:i])
            ancestors.append(ancestor)
        return ancestors

    def get_ancestor_summaries(self, relative_path: str) -> Dict[str, str]:
        """Given a relative_path, construct a Dict with key is ancestor path and value is the summary"""
        ancestor_summaries = {}
        ancestors = self.get_ancestors(relative_path)
        for ancestor in ancestors:
            if ancestor in self.directory_summaries:
                ancestor_summaries[ancestor] = self.directory_summaries[ancestor]
        return ancestor_summaries

    def summarize_file(self, absolute_path: str, relative_path: str, verbose: bool = False) -> str:
        """
        absolute_path: for opening file
        relative_path: for getting ancestor directory summaries
        """
        try:
            with open(absolute_path, "r", encoding="utf-8") as f:
                file_contents = f.read()
        except (UnicodeDecodeError, IOError) as e:
            logger.warning(f"Error reading file {relative_path}: {str(e)}")
            # if the file can't be opened (e.g. it's an image), use the file name as content
            file_contents = os.path.basename(absolute_path)

        if len(file_contents.splitlines()) > 10000:  # Skip large files
            return ""

        file_hash = hash_code(file_contents)

        if (results := file_summary_cache.get(file_hash)) is not None:
            return results

        # TODO: use older file contents in dev
        if DEV:
            try:
                commits = self.cloned_repo.git_repo.git.log(relative_path, n=10, format="%H").splitlines()
                for commit in commits:
                    try:
                        previous_file_contents = self.cloned_repo.git_repo.git.show(f"{commit}:{relative_path}")
                        if (results := file_summary_cache.get(hash_code(previous_file_contents))) is not None:
                            return results
                    except Exception as e:
                        if "exists on disk, but not in" in str(e):
                            # File didn't exist in this commit, skip to the next one
                            continue
                        else:
                            logger.warning(f"Error getting previous file contents for {relative_path}: {str(e)}")
            except Exception as e:
                logger.warning(f"Error getting previous file contents for {relative_path}: {str(e)}")

        ancestor_summaries = self.get_ancestor_summaries(relative_path)

        directory_summaries = ""
        for ancestor, summary in ancestor_summaries.items():
            directory_summaries += f"{ancestor}\n{summary}\n\n"
        directory_summaries = directory_summaries.strip()

        response = call_llm(
            system_prompt,
            user_prompt,
            params={
                "DIRECTORY_SUMMARIES": directory_summaries,
                "FILE_NAME": relative_path,
                "FILE_CONTENTS": file_contents,
            },
            verbose=False,
        )
        summary = extract_xml_tag(response, "file_summary")
        file_summary_cache[file_hash] = summary
        return summary

    def generate_file_summaries(self):
        """Walk through the cloned repo and generate summaries for all code files in parallel"""
        file_summaries = {}
        sweep_config = SweepConfig()
        files_to_summarize = []

        # Collect all files to summarize
        for root, dirs, files in os.walk(self.cloned_repo.repo_dir):
            dirs[:] = [d for d in dirs if d not in sweep_config.exclude_dirs]
            for file in files:
                absolute_path = os.path.join(root, file)
                relative_path = os.path.relpath(absolute_path, self.cloned_repo.repo_dir)
                if is_actual_code_file(root, absolute_path, sweep_config):
                    files_to_summarize.append((absolute_path, relative_path))

        def summarize_wrapper(args):
            abs_path, rel_path = args
            try:
                return rel_path, self.summarize_file(abs_path, rel_path)
            except Exception as e:
                logger.error(f"Error summarizing {rel_path}: {str(e)}")
                return rel_path, None

        workers = min(4, len(files_to_summarize)) if DEV else 1
        # WARNING: this is unsafe because it can trigger an ValuError: I/O operation on closed file
        if files_to_summarize:
            results = joblib.Parallel(n_jobs=workers, prefer="threads")(
                joblib.delayed(summarize_wrapper)(args)
                for args in tqdm(files_to_summarize, desc="Summarizing files", file=sys.stderr)
            )

            file_summaries.update({rel_path: summary for rel_path, summary in results if summary is not None})

        return file_summaries


if __name__ == "__main__":
    from sweepai.core.github_utils import github_integration

    REPO = os.environ["REPO"]
    org_name, repo_name = REPO.split("/")
    installation_id = github_integration.get_repo_installation(org_name, repo_name)
    cloned_repo = ClonedRepo(REPO, installation_id=installation_id)
    file_summarizer = FilesSummarizer(cloned_repo=cloned_repo)
    # print(file_summarizer.file_summaries)



import json
import os
from dataclasses import dataclass
from datetime import datetime
from typing import Any

import requests
from loguru import logger

from sweepai.config.server import CACHE_DIRECTORY, POSTHOG_API_KEY
from sweepai.o11y.telemetry_utils import make_serializable_dict

POSTHOG_EVENTS_LOCATION = f"{CACHE_DIRECTORY}/posthog_events.txt"


def write_posthog_event_to_disk(event: str, properties: dict[str, Any] | None = None):
    try:
        properties["event"] = event
        full_payload = properties | {
            "event": event,
            "timestamp": datetime.utcnow().isoformat() + "Z",
        }
        stringified_payload = json.dumps(make_serializable_dict(full_payload)) + "\n"
    except Exception as e:
        logger.error(f"Error stringifying posthog event: {e}")
        return
    fd = os.open(POSTHOG_EVENTS_LOCATION, os.O_WRONLY | os.O_APPEND | os.O_CREAT)
    try:
        os.write(fd, stringified_payload.encode())
    except Exception as e:
        logger.error(f"Error writing posthog event to disk: {e}")
        os.close(fd)
        return
    finally:
        os.close(fd)


@dataclass
class PosthogClient:
    """
    Official Posthog API client has a thread leakage, so we are using a custom client.
    """

    API_KEY: str | None = None

    def capture(
        self,
        distinct_id: str | None = None,
        event: str | None = None,
        properties: dict[str, Any] | None = None,
        **kwargs,
    ):
        url = "https://app.posthog.com/capture/"
        headers = {"Content-Type": "application/json"}
        payload = {
            "api_key": self.API_KEY,
            "event": event,
            "properties": {"distinct_id": distinct_id, **properties},
            "timestamp": datetime.utcnow().isoformat() + "Z",  # Adding 'Z' to indicate UTC time
        }
        # always write to disk
        write_posthog_event_to_disk(event=event, properties=properties)
        if self.API_KEY is None:
            return  # don't fire off to posthog if we don't have an API key
        else:
            response = requests.post(url, headers=headers, data=json.dumps(make_serializable_dict(payload)))
            return response


posthog = PosthogClient(API_KEY=POSTHOG_API_KEY)



system_prompt = """You are a careful and smart tech lead that wants to avoid production issues. You will be analyzing a set of diffs representing a pull request made to a piece of source code. 
You will also be given the pull request title and description which you will use to determine the intentions of the pull request. Be very concise."""

system_prompt_review = """You are a busy tech manager who is responsible for reviewing prs and identifying any possible production issues. 
You will be analyzing a list of potential issues that have been identified by a previous engineer and determing which issues are severe enough to bring up to the original engineer."""

system_prompt_special_rules = """You are a careful and smart tech lead that wants to avoid production issues. You will be analyzing a set of diffs representing a pull request made to a piece of source code. 
You will also be given the pull request title and description which you will use to determine the intentions of the pull request. Finally you will be given a set of rules to check the pull request changes against.
It is your job to make sure that the pull request changes do not violate any of the given rules.
"""

system_prompt_edge_case_question_creation = """You are a careful and smart tech lead that wants to avoid production issues. You will be analyzing a set of diffs representing a pull request made to a piece of source code.
It is your job to ask a series of detailed and exhaustive set of questions which will then be passed on to the original engineer. The point of this is to make sure that after all the questions have been answered, it will be clear whether or not there are any production issues with the pull request. You are to tailor the questions you create to the code changes provided.
"""

system_prompt_edge_case_answer = """You are a careful and smart tech lead that wants to avoid production issues. You will be analyzing a set of diffs representing a pull request made to a piece of source code.
It is your job to answer a series of questions presented to you by a previous engineer. The questions are meant to help identify any potential production issues with the pull request. You are to answer these questions and using those answers determine if there are any production issues with the pull request.
"""

system_prompt_identify_new_functions = """You are an expert programmer with a keen eye for detail, assigned to analyze a series of code patches in a pull request. Your primary responsibility is to meticulously identify all newly created functions within the code."""

system_prompt_identify_repeats = """You are a proficient programmer tasked with identifying useless utility functions in a codebase. Your job is to identify any useless function definitions.
You will be given a function definition that was just added to the codebase and your job will be to check whether or not this function was actually necessary or not given a series of related code snippets.
"""

system_prompt_pr_summary = """You are a talented software engineer that excels at summarising pull requests for readability and brevity. You will be analysing a series of patches that represent all the changes made in a specific pull request.
It is your job to write a short and concise but still descriptive summary that describes what the pull request accomplishes."""

system_prompt_sort_issues = """You are a helpful and detail-oriented software engineer who is responsible for sorting a list of identified issues based on their severity and importance. 
You will be analyzing a list of issues that have been identified by a previous engineer and determing the severity of each issue.
You will then rank the issues from most severe to least severe based on your analysis."""

user_prompt = """\
# Pull Request Title and Description
Here is the title and description of the pull request, use this to determine the intentions of the pull request:

{pull_request_info}

# Code Review
Here are the changes in the pull request changes given in diff format:
<changes>
{diff}
</changes>

# Instructions
1. Analyze the code changes. Keep in mind what the intentions of the pull request are.
    1a. Review each change individually, examining the code changes line-by-line.
    1b. For each line of code changed, consider:
        - What is the purpose of this line of code?
        - How does this line of code interact with or impact the rest of the files in the pull request?
        - Is this line of code functionally correct? Could it introduce any bugs or errors?
        - Is this line of code necessary? Or could it be an accidental change or commented out code?
    1c. Describe all changes that were made in the diffs. Respond in the following format. (1 paragraph)
<thoughts>
<thinking>
{{Analysis of change 1, include all questions and answers}}
</thinking>
...
</thoughts>
    1d. Provide a final summary for the changes that should be a single sentence and formatted within a <change_summary> tag.
Here is an example, make sure the summary sounds natural and keep it brief and easy to skim over:
<example_change_summary>
Added a new categorization system for snippets in `multi_prep_snippets` and updated the snippet score calculation in `get_pointwise_reranked_snippet_scores`. 
</example_change_summary>
<change_summary>
{{Final summary of the major changes}}
</change_summary>

2. Identify all issues.
    2a. Determine whether there are any functional issues, logical errors or runtime errors that the code changes introduce. Consider the line-by-line analysis from step 1b. (1 paragraph)
    2b. Only include issues that you are very confident will cause serious issues that prevent the pull request from being merged. For example, focus only on functional code changes and ignore changes to strings and comments that are purely descriptive.
    2c. Do not make assumptions about existing functions or code. Assume all existing code and system configurations are correct and functioning as intended.
    2d. Do not raise issues over functions being called without error handling. You do not have enough context to determine if the functions are being called correctly.
    2e. You must take into account the intentions of the pull request when identifying issues. Are the the code changes in line with the intentions of the pull request? If the answer is yes then this is not an issue.

Answer each of the above questions in step 2 in the following format:
<issue_identification>
{{Include all answers to the question posed in 2a, to 2f. Example: 2a: A potential...}}
</issue_identification>
"""
user_prompt_special_rules_format = """
# Code Review
Here are the changes in the pull request changes given in diff format:
<changes>
{diff}
</changes>

# Rules
Here are list of rules that are specific to this code review:
<rules>
{special_rules}
</rules>

# Instructions
Along with the rules provided, there may be examples given for each rule. These examples are important to consider when analyzing the code changes.
1. Analyze the rules and examples provided.
    For each rule provided, answer the following:
    Rule #n:
        a. Are there any examples relating to this rule? If no, you may stop here. If yes, repeat the example verbatim here.
        b. For each example, examine if the example provided relates to any code change in the pull request. If it does, explain why, if it doesn't explain why.
Output the questions and answers for each rule in step 1 in the following format:
<examples_analysis>
{{Question and answers for each example in the special_rules section.}}
</examples_analysis> 

2. Analyze the code changes.
    For each rule provided, answer the following questions:
    Rule #n:
        a. Are there code changes in violation of the rule? If yes, then this is an issue that should be raised.
        b. Are there any examples for this rule? If yes, then for each example provided explicitly list out the example and check if the rule applies. Remember that the examples are there for a reason!
Output the questions and answers for each rule in step 2 in the following format:
<rules_analysis>
{{Question and answers for each rule in the special_rules section.}}
</rules_analysis>
"""

user_prompt_edge_case_question_creation_format = """
# Code Review
Here are the changes in the pull request changes given in diff format:
<changes>
{diff}
</changes>

# Instructions
1. Summarize the code changes
Review each change individually, examining the code changes line-by-line. Then review all the changes at once in order to get an idea of how each individual change combines together and what the overall effect of the pull request is.

Respond in the format below, this is mandatory:
<thoughts>
{{Provide your thoughts on the code changes above, providing a summary and how each change ties together with the other changes.}}
</thoughts>

2. Create the questions
You will be given following Cases in which you much generate a series of questions. For each case you will be given guidelines on how to create these questions most effectively. You are to tailor the questions you create to the code changes provided. Make sure to explicitly reference variable names and entities related to the questions and even line numbers if necessary. You MUST create the questions in a way such that an affirmative answer "yes" means that there is an issue with the pull request. That means you should phrase you questions in a way such that they are asking if this issue exists or not. ALWAYS begin the question by asking "Yes or No?"
<cases>
<case_1>
Case 1 - Create questions related to concurrency changes. If there are changes related to concurrency in the pull request you are to generate questions related to concurrency changes.  If not, you may skip this part and go to the next Case. The questions you generate should be an exhaustive list of questions meant to help identify issues with the code changes.
<guidelines_for_case_1>
1. Create questions asking if the variables are mutated unintentionally in any way.
2. When there are variables such as arrays where the order of the data matters, ask questions about whether or not the order of these variables are mutated.
3. Be strict with how you ask your questions. Do not use words like "potentially" or "possibly". If you are asking if an issue exists, ask in a direct manner. Example: Instead of asking "Is there any chance that there is a deadlock" instead ask "Is there a deadlock".
4. Phrase the questions to make sure that if the answer is "Yes" then that means there is an issue with the pull request. Basically you are asking if an issue exists or not.
</guidelines_for_case_1>
</case_1>
</cases>

Respond with ALL your created questions for every case in the following format below, this is mandatory:
<created_questions>
1. Yes or No? Question 1 goes here...
2. Yes or No? Question 2 goes here...
...
n. Yes or No? Question n goes here...
</created_questions>

If there are no relevant changes simply respond with the following:
# No relevant changes detected
"""

user_prompt_edge_case_question_answer_format = """
# Code Review
Here are the changes in the pull request changes given in diff format:
<changes>
{diff}
</changes>

# Questions to Answer
<questions_to_answer>
{questions}
</questions_to_answer>

# Instructions
1. Summarize the code changes
Review each change individually, examining the code changes line-by-line. Then review all the changes at once in order to get an idea of how each individual change combines together and what the overall effect of the pull request is.

Respond in the format below, this is mandatory:
<thoughts>
{{Provide your thoughts on the code changes above, providing a summary and how each change ties together with the other changes.}}
</thoughts>

2. Answer the # Questions to Answer
Now you must answer each question, justifying each answer as you go and giving an example from the code changes to support your answer.
First repeat the question verbatim then answer the question in the following format below, this is mandatory:
<questions_and_answers>
1. Question 1 repeated verbatim. Answer: Question 1 answer goes here...
2. Question 2 repeated verbatim. Answer: Question 2 answer goes here...
...
n. Question n repeated verbatim. Answer: Question n answer goes here...
</questions_and_answers>

3. Identify Issues
Based on your answers to the questions above you are now to identify any potential issues that the code changes may introduce. 
Read through each question and answer pair and determine if an issue needs to be raised. 
If there was not enough information to answer a question IT IS NOT AN ISSUE! DO NOT INCLUDE IT AS AN ISSUE AND MOVE ON TO THE NEXT QUESTION. 
"""

user_prompt_issue_output_format = """
[FORMAT]
Finally, format the found issues and root causes using the following XML tags. Each issue description should be a single sentence. 
Include a corresponding line number for the issue. The issue will be raised as a github comment on that exact line in the code file. DO NOT reference the patch or patch number in the description. Format these fields in an <issue> tag in the following manner:

<issues>
<issue>
<issue_description>
{{Issue 1 description, be sure to reference the line/lines of code. Max 1 sentence.}}
</issue_description>
<file_name>
{{Corresponding file name that this issue is in}}
</file_name>
<line_number>
{{Corresponding line number for Issue 1. This is the line the github comment will be on when the issue is raised.}}
</line_number>
</issue>
...
</issues>
If there are no issues found do not include the corresponding <issue> tag.

Focus your analysis solely on potential functional issues with the code changes. Do not comment on stylistic, formatting, or other more subjective aspects of the code."""

user_prompt_review_questions = """
Below are a series of identified issues for the following files {file_names} formatted in the following way:
<potential_issues>
<issue>
<issue_description>
{{Issue 1 description}}
</issue_description>
<file_name>
{{Corresponding file name for Issue 1}}
</file_name>
<line_number>
{{Corresponding line number for Issue 1}}
</line_number>
</issue>
...
</potential_issues>

# Potential Issues

{potential_issues}

Below are the changes made in the pull request as context
# Relevant code files with line numbers

{pr_changes}

Below is the title and description of the pull request. Use this information to determine the intentions of the pull request and then further scrutinize the potential issues.
# Pull Request Title and Description

{pull_request_info}

Below is a summary of the pull request. Use this to determine the intentions of the pull request to check if the potential issues are actually intended changes.
# Pull Request Summary

{pull_request_summary}

Below are the comments left on the pull request, possibly from a previous review you did. Use these comments to determine if there are any issues that have already been identified or if there is anything you need to pay attention to.
The issues that have already been identified should not be raised again.
# Previous Review and Comments left on the Pull Request

{comment_threads}

# Instructions
1. Analyze each identified potential issue for the file(s) {file_names}
Review each identified issue individually. Answer all of the below questions for every potential issue identified.
    1a. Is this reported issue accurate (double check that the previous reviewer was not mistaken, YOU MUST include the corresponding patch for proof). If the answer to this question is no, then the issue is not severe. 
    1b. Is there accidentally removed or commented out lines of code that has functional utility. In this case double check if this change was intentional or accidental.
    1c. Take into account the intentions of the pull request when identifying issues. Are the code changes in line with the intentions of the pull request? If the answer is yes then this is not an issue.
    1d. Is this issue related to potential security vulnerabilities? If yes, then this issue is NOT severe and should not be included.
    1e. Does this issue have the potential to cause a production crash or introduce a logical error? If yes, then this issue is severe and should be included in the final list of issues.
    1f. Is this issue actually an intended change based off of the pull request title and summary? If the issue relates to any code change mentioned in the pull request summary then the answer to this question is yes. If the answer is yes, do NOT include this a severe issue.
    1g. Finally was this issue already raised in a comment thread? If yes, then this issue has already been identified and you should not raise it again. You must also provide proof by referencing the exact comment where this issue was raised.
    1h. Is this issue severe based on the questions and answers above? Any issue that will cause production to crash or introduce a logical error is considered severe.
"""
user_prompt_review_special_rules = """
In addition to all the above questions you must answer in step 1, the following rules also apply to this file as defined by the user themselves:

<special_rules>
{special_rules}
</special_rules>

For each rule defined in the special_rules section, ask if the issue is in violation of the rule. 
You may be given relevant context for a rule in which case extra attention is required to make sure that the change in the pull request does not violate the rule.
If the issue is in violation of the rule, then it is severe and should be included in the final list of issues.
"""

user_prompt_review_analysis_format = """    
Deliver your analysis including all questions and answers in the following format:
<thoughts>
<thinking>
{{Analysis of each issue, include ALL the questions and answers.
Answer in this format for all issues: 
Issue 1: Issue 1 description
1a. Answer for question 1a...
1b. ...
...
1h. Answer for question 1h...}}
</thinking>
...
</thoughts>"""

user_prompt_review_decisions = """
2. Decide which issues to keep
    2a. Based on your analysis in step 1, now decide which issues to keep and drop. Only include severe issues. To determine if an issue is severe examine the questions and answers you provided above.
    2b. Remember that an issue is severe if and ONLY if it has the potential to cause a production crash or introduce a logical error. If the issue is not severe then it should be dropped.
    2c. After choosing to keep an issue you are to respond in the following format:
<severe_issues>
<issue>
<issue_description>
{{Issue 1 description}}
</issue_description>
<file_name>
{{Corresponding file name for Issue 1}}
</file_name>
<line_number>
{{Corresponding line number for Issue 1}}
</line_number>
</issue>
...
</severe_issues>
"""

user_prompt_identify_new_functions = """Below are all the patches made to the file {file_name} in this pull request. Use these patches to determine if there are any newly created utility functions.
# PR Patches

{patches}

Below is the file {file_name} with all the above patches applied along with the line numbers. Use this to identify the correct starting and ending line numbers.
# Relevant code file with line numbers

{numbered_code_file}

# Instructions
1. Analyze each of the patches above and identify any newly created utility functions. To determine if a function is newly created, answer the following:
    1a. Note that if a function is renamed such as having of its parameters changed, or if a function has been reworked meaning the contents of the file has changed, this should not be included as a newly created function.
    1b. Is the function created from scratch? If not, the function is not newly created.
    1c. Is there a corresponding patch that shows the creation of the function? If the answer is no, then the function is not newly created. If the answer is yes, give the patch number as proof.
    1d. Is this function a utility function? It should be relatively short and simple, and should not be a class method. If it is too long and complex then do not include it in the final list.
    1e. Answer these questions in the following xml format:
<thinking>
{{Questions and answer for each function that you believe is newly created.}}
</thinking>
2. Based on the questions and answers above return the list of all newly created utility functions in the following xml format:
<newly_created_functions>
<function>
<function_code>
{{Function code copied verbatim for the patch}}
</function_code>
<start_line>
{{Corresponding starting line number for function 1 (inclusive)}}
</start_line>
<end_line>
{{Corresponding ending line number for function 1 (inclusive)}}
</end_line>
</function>
...
</newly_created_functions>
"""

user_prompt_identify_repeats = """
Below is the pull request title and description. Use this to determine the intention of the pull request to help determine if the new function is useless or not.
# Pull Request Title and Description

{pull_request_info}

Below is the function definition that was just added to the code base.
# New Function

{function}

Below are a series of code snippets retrieved from the codebase via vector search. Analyze these code snippets to see if there are any functions that fulfill the exact same purpose and has the same input and output which would render the new function useless.
# Relevant code snippets

{formatted_code_snippets}

Below are the comments left on the pull request, possibly from a previous review you did. Use these comments to determine if there are any issues that have already been identified or if there is anything you need to pay attention to.
The issues that have already been identified should not be raised again.
# Previous Review and Comments left on the Pull Request

{comment_threads}

# Instructions
1. Analyze each of the code snippets above and determine whether or not the new function is useless. Specifically, compare the new function with the existing methods in the code snippets by answering ALL the following questions:
   1a. Purpose: What is the primary purpose of the new function? Is this purpose already served by an existing method? If its purpose is not perfectly serve by an existing method then this function is not useless. If it takes more than one existing function to replicate the functionality of this new function, than this new funciton is not useless.
   1b. Intention: What was the intention behind adding this new function? Is this function meant to be a wrapper or an interface function? If the answer is yes, then this new function is important and should not be removed.
   1c. Functionality: What specific tasks or operations does the new function perform? Are these tasks or operations already handled by existing methods?
   1d. Initialization: What data structures or variables are initialized in the new function? Are similar initializations present in existing methods?
   1e. Data Processing: How does the new function process data (e.g., formatting, extracting, or transforming data)? Are these data processing steps already implemented in existing methods?
   1f. Unique Contributions: Does the new function provide any unique contributions or improvements that are not covered by existing methods? If it does then it should be considered as not useless and should be kept.
   1g. Impact of Removal: Would removing this function require a significant refactor of existing functions? Would the use cases of the existing functions change at all? If the answer is yes to any of these questions the new function is not useless.
   1h. Already Mentioned: Has this issue been previously raised in a comment thread? If yes, then this function is considered not useless. Provide proof of it being raised by referencing the exact comment where this issue was raised.

2. Return your answer in the following xml format:
<useless_new_function>
<thinking>
{{Any thoughts/analysis you have should go here. This is where you MUST answer each of the questions above.}}
</thinking>
<answer>
{{'true' if the new function is useless, 'false' if the new function provides unique contributions.}}
</answer>
<justification>
{{A very brief justification of the decision made. When justifying why make sure to reference relevant functions. Max 1-2 sentences.}}
</justification>
</useless_new_function>"""

user_prompt_pr_summary = """
Below is the title and description of the pull request. Use this to determine the intentions of the pull request.

{pull_request_info}

Below are all the patches associated with this pull request along with each of their file names

# All Pull Request Patches

{all_patches}

1. Summarise the major changes using the following principles:
    1a. Begin with a 2-3 line overall summary of what the main goal of the pull request was.
    1b. Describe any major or noteworthy implementation details with a short concise 1 sentence.
    1c. Never provide "useless" summaries. "useless" summaries are the following: informing the user a variable or function was created without explaining how it contributes the the main goal of the pull request.
    1d. Instead summarize how the changes were accomplished like this: function `foo` implements feature bar and this had xyz effect of abc.
    1e. Do not summarize minor changes that do not tie into the main goal of the pull request.
    1f. Avoid using overly complex language. For example: instead of the word 'utilize' instead use the word 'use'. 
    1g. Respond in the following xml format:
<preamble>
Say "I will keep my summary short and concise. I will sounds business-like and get straight to the point."
</preamble>
<pr_summary>
{{Provide a detailed summary here. Be sure to reference relevant entities and variables to make it very clear what you are referencing. Speak in past tense. 
This summary should be maximum 10 sentences. Make sure the summary is not a wall of text, use an adequate amount of new lines.}}
</pr_summary>

Here are a few example <pr_summary></pr_summary> blocks:
<example_preamble>
I will keep my summary short and concise. I will sounds business-like and get straight to the point.
</example_preamble>
<example_pr_summary>
Added support for bulk actions to the admin dashboard.

A new `BulkActionDropdown` component in `components/BulkActionDropdown.vue` was created that renders a dropdown menu with options for bulk actions that can be performed on selected items in the admin dashboard.

The existing `ItemList` component in `components/ItemList.vue` was updated to include checkboxes for each item and to enable the `BulkActionDropdown` when one or more items are selected. 

A new `bulkDelete` action was added to the item store in `store/item.js` which accepts an array of item IDs and deletes them from the database in a single query. The `BulkActionDropdown` component dispatches this action when the "Delete Selected" option is chosen.

Unit tests were added for the `BulkActionDropdown` component and the `bulkDelete` store action in the `components/BulkActionDropdown.test.js` and `store/item.test.js` files respectively.
</example_pr_summary>

<example_preamble>
I will keep my summary short and concise. I will sounds business-like and get straight to the point.
</example_preamble>
<example_pr_summary>
Added two factor authentication to the user authentication process.

The `loginUser` function in `handlers/auth.js` now calls the new `verifyTwoFactorCode` function located in `utils/auth-utils.js` after validating the user's password. `verifyTwoFactorCode` is responsible for verifying the user's two-factor authentication code.

Unit tests were added for `verifyTwoFactorCode` in `tests/auth.test.js` covering the following scenarios: providing a valid code, an expired code, and an invalid code.

Additionally, the documentation in `README.md` was updated to reflect the changes to the authentication flow and now describe the new two-factor authentication step.
</example_pr_summary>
"""

user_prompt_sort_issues = """Below are all the identified issues for the pull request. You will be sorting these issues based on severity and importance.

# Identified Issues

{all_issues}

# Instructions
1. Review each issue and determine the severity of the issue.
    1a. Output your analysis in the following format:
<thinking>
{{Analysis of each issue, include the estimated severity and a reason as to why}}
</thinking>
2. Rank the issues based on severity and importance.
    2a. Based on your analysis in step 1, rank the issues from most severe to least severe.
    2b. You must respond with the the following xml format, returning a list of the issue indices in order of severity:

<sorted_issue_indices_by_severity>
{{Sorted indices go here, example: 1, 3, 4 ,2}}
</sorted_issue_indices_by_severity>
"""


pr_changes_prefix = "The following changes were made in the PR. Each change contains all of the patches that were applied to a file, as well as the source code after the change.\n"

pr_review_change_group_format = """
Below are a series of patches for the following files: {file_names}
# All Patches

<all_patches>
{all_patches}
</all_patches>

# Below are the source code files for the following files {file_names} with the above patches applied

<all_source_code>
{all_source_code}
</all_source_code>
"""

source_code_with_patches_format = """
Here is the source code for file {file_name} after applying all patches:
<source_code file_name="{file_name}">
{file_contents}
</source_code>"""

patches_for_file_format = """
<file_patches file_name="{file_name}">
{file_patches}
</file_patches>
"""

patch_format = """\
<patch file_name="{file_name}" index="{index}">
{diff}
</patch>
<patch_annotation file_name="{file_name}" index="{index}">
{annotation}
</patch_annotation>"""

patch_format_without_annotations = """\
<patch file_name="{file_name}" index="{index}">
{diff}
</patch>"""

comment_thread_format = """
<comment_thread file_name="{file_name}" line_number="{line_number}" is_resolved="{is_resolved}">
{comments}
</comment_thread>
"""

comment_format = """<comment file_name="{file_name}" line_number="{line_number}">
{comment_body}
</comment>"""

comment_thread_is_resolved_format = """
The following comment thread has been RESOLVED!
This means the issue you raised in this comment thread is no longer an issue. Do not raise this issue again.
{comment_thread}
"""



"""This file contains functions to read and display posthog events from disk as html."""

import json

from sweepai.o11y.event_logger import POSTHOG_EVENTS_LOCATION


def read_posthog_events_from_disk() -> list[str]:
    """Reads posthog events from disk."""
    with open(POSTHOG_EVENTS_LOCATION, "r") as f:
        return f.readlines()


def format_posthog_events_as_csv(events: list[str]) -> str:
    """Format posthog events as a csv"""
    csv = "username,timestamp,event\n"
    for event in events:
        event_dict = json.loads(event)
        username = event_dict.get("username", "Unknown User")
        event_name = event_dict.get("event", "Unknown Event")
        timestamp = event_dict.get("timestamp", "Unknown Timestamp")
        csv += f"{username},{timestamp},{event_name}\n"
    return csv


if __name__ == "__main__":
    events = read_posthog_events_from_disk()
    csv = format_posthog_events_as_csv(events)
    with open("posthog_events.csv", "w") as f:
        f.write(csv)



from __future__ import annotations

import copy
import json
import os
import re
import shutil
import subprocess
import tempfile
import time
import traceback
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime, timedelta
from functools import cached_property

import git
import requests
from github import (
    Github,
    GithubException,
    GithubIntegration,
    InputGitTreeElement,
    PullRequest,
)
from github.Auth import AppAuth, Token
from github.GithubException import BadCredentialsException, UnknownObjectException
from github.Repository import Repository

# get default_base_url from github
from github.Requester import Requester
from loguru import logger
from urllib3 import Retry

from sweepai.config.client import SweepConfig
from sweepai.config.server import (
    CACHE_DIRECTORY,
    CURL_CA_BUNDLE,
    GITHUB_API_BASE_URL,
    GITHUB_APP_ID,
    GITHUB_APP_PEM,
    GITHUB_BASE_URL,
)
from sweepai.core.entities import FileChangeRequest
from sweepai.o11y.log_utils import suppress_with_warning
from sweepai.utils.str_utils import get_hash

MAX_FILE_COUNT = 50

git_env = {"GIT_SSL_CAINFO": CURL_CA_BUNDLE} if CURL_CA_BUNDLE else {}

all_repos: list[ClonedRepo] = []  # list of all ClonedRepo objects

auth = AppAuth(app_id=GITHUB_APP_ID, private_key=GITHUB_APP_PEM) if GITHUB_APP_ID and GITHUB_APP_PEM else None
github_integration = GithubIntegration(auth=auth, base_url=GITHUB_API_BASE_URL) if auth else None


def make_valid_string(string: str):
    pattern = r"[^\w./-]+"
    return re.sub(pattern, "_", string)


def get_jwt(signing_key: str = "", app_id: str = ""):
    from jwt import encode

    if not signing_key:
        signing_key = GITHUB_APP_PEM
    if not app_id:
        app_id = GITHUB_APP_ID
    # max is 10 minutes, use 5 minutes to be very safe. 9min 50sec failed before
    payload = {"iat": int(time.time()), "exp": int(time.time()) + 300, "iss": app_id}
    return encode(payload, signing_key, algorithm="RS256")


def get_token(installation_id: int, signing_key: str = "", app_id: str = ""):
    if int(installation_id) < 0:
        logger.warning(f"installation_id is {installation_id}, using GITHUB_PAT instead.")
        return os.environ["GITHUB_PAT"]
    for timeout in [5.5, 5.5, 10.5]:
        try:
            jwt = get_jwt(signing_key=signing_key, app_id=app_id)
            headers = {
                "Accept": "application/vnd.github+json",
                "Authorization": "Bearer " + jwt,
                "X-GitHub-Api-Version": "2022-11-28",
            }
            response = requests.post(
                f"{GITHUB_API_BASE_URL}/app/installations/{int(installation_id)}/access_tokens",
                headers=headers,
            )
            obj = response.json()
            if "token" not in obj:
                logger.error(obj)
                raise Exception("Could not get token")
            return obj["token"]
        except SystemExit:
            raise SystemExit
        except Exception:
            time.sleep(timeout)
    signing_key = GITHUB_APP_PEM
    app_id = GITHUB_APP_ID
    exception_message = "Could not get token."
    if not signing_key:
        exception_message += " Missing GITHUB_APP_PEM in the .env file."
    if not app_id:
        exception_message += " Missing GITHUB_APP_ID in the .env file."
    if signing_key and app_id:
        exception_message += "Please double check that Sweep has the correct permissions to access your repository."

    raise Exception(exception_message)


def get_app(signing_key: str = "", app_id: str = ""):
    jwt = get_jwt(signing_key=signing_key, app_id=app_id)
    headers = {
        "Accept": "application/vnd.github+json",
        "Authorization": "Bearer " + jwt,
        "X-GitHub-Api-Version": "2022-11-28",
    }
    response = requests.get(f"{GITHUB_API_BASE_URL}/app", headers=headers)
    return response.json()


def get_github(repo_full_name: str):
    org_name, repo_name = repo_full_name.split("/")
    return CustomGithub(installation_id=get_installation_id(username=org_name), base_url=GITHUB_API_BASE_URL)


def get_github_repo_object(repo_full_name: str):
    g = get_github(repo_full_name)
    return g.get_repo(repo_full_name)


class CustomRequester(Requester):
    def __init__(
        self,
        token,
        timeout: int = 15,
        user_agent: str = "PyGithub/Python",
        per_page: int = 30,
        verify: bool = True,
        retry=None,
        pool_size=None,
        installation_id: int = None,
    ) -> "CustomRequester":
        self.token = token
        self.installation_id = installation_id
        base_url = GITHUB_API_BASE_URL
        auth = Token(token)
        retry = Retry(
            total=3,
        )  # 3 retries
        self.timeout = timeout
        self.user_agent = user_agent
        self.per_page = per_page
        self.verify = verify
        self.retry = retry
        self.pool_size = pool_size
        super().__init__(
            auth=auth,
            base_url=base_url,
            timeout=timeout,
            user_agent=user_agent,
            per_page=per_page,
            verify=verify,
            retry=retry,
            pool_size=pool_size,
        )

    def _refresh_token(self):
        """Get a new token and reinitialize the requester with it"""
        try:
            self.token = github_integration.get_access_token(self.installation_id).token
        except Exception:
            logger.warning("[YELLOW] Could not get access token from GitHub API using PyGithub, using legacy method")
            self.token = get_token(self.installation_id, signing_key=GITHUB_APP_PEM, app_id=GITHUB_APP_ID)
        
        # Reinitialize with new token
        auth = Token(self.token)
        super().__init__(
            auth=auth,
            base_url=GITHUB_API_BASE_URL,
            timeout=self.timeout,
            user_agent=self.user_agent,
            per_page=self.per_page,
            verify=self.verify,
            retry=self.retry,
            pool_size=self.pool_size,
        )

    def requestJsonAndCheck(self, *args, **kwargs):
        try:
            return super().requestJsonAndCheck(*args, **kwargs)
        except (BadCredentialsException, UnknownObjectException) as e:
            if isinstance(e, BadCredentialsException) or (
                isinstance(e, UnknownObjectException) and e.status in (401, 403)
            ):
                self._refresh_token()
                return super().requestJsonAndCheck(*args, **kwargs)
            raise e


class CustomGithub(Github):
    def __init__(
        self,
        installation_id: int,
        base_url: str = GITHUB_API_BASE_URL,
        *args,
        **kwargs,
    ) -> "CustomGithub":
        self.installation_id = installation_id
        self.token = self._get_token()
        
        # Create our custom requester first
        requester = CustomRequester(
            self.token, 
            installation_id=self.installation_id,
        )
        
        # Initialize with the requester directly
        self._Github__requester = requester

    @classmethod
    def from_repo_full_name(cls, repo_full_name: str, **kwargs):
        org_name, repo_name = repo_full_name.split("/")
        try:
            installation_id = github_integration.get_repo_installation(org_name, repo_name).id
        except Exception:
            logger.warning("[YELLOW] Could not get access token from GitHub API using PyGithub, using legacy method")
            installation_id = get_installation_id(org_name)
        return cls(installation_id, **kwargs)

    def _get_token(self) -> str:
        try:
            return github_integration.get_access_token(self.installation_id).token
        except Exception:
            logger.warning("[YELLOW] Could not get access token from GitHub API using PyGithub, using legacy method")
            return get_token(self.installation_id, signing_key=GITHUB_APP_PEM, app_id=GITHUB_APP_ID)

def get_github_client(
    installation_id: int, 
) -> tuple[str, CustomGithub]:
    github_instance = CustomGithub(
        installation_id,
    )
    return github_instance.token, github_instance


# fetch installation object
def get_installation(username: str, signing_key: str = "", app_id: str = ""):
    jwt = get_jwt(signing_key=signing_key, app_id=app_id)
    try:
        # Try user
        response = requests.get(
            f"{GITHUB_API_BASE_URL}/users/{username}/installation",
            headers={
                "Accept": "application/vnd.github+json",
                "Authorization": "Bearer " + jwt,
                "X-GitHub-Api-Version": "2022-11-28",
            },
        )
        obj = response.json()
        return obj
    except Exception:
        # Try org
        response = requests.get(
            f"{GITHUB_API_BASE_URL}/orgs/{username}/installation",
            headers={
                "Accept": "application/vnd.github+json",
                "Authorization": "Bearer " + jwt,
                "X-GitHub-Api-Version": "2022-11-28",
            },
        )
        try:
            obj = response.json()
            return obj["id"]
        except Exception as e:
            logger.error(e)
            logger.error(response.text)
        raise Exception("Could not get installation, probably not installed")


def get_installation_id(username: str) -> int:
    jwt = get_jwt(signing_key=GITHUB_APP_PEM, app_id=GITHUB_APP_ID)
    try:
        # Try user
        response = requests.get(
            f"{GITHUB_API_BASE_URL}/users/{username}/installation",
            headers={
                "Accept": "application/vnd.github+json",
                "Authorization": "Bearer " + jwt,
                "X-GitHub-Api-Version": "2022-11-28",
            },
        )
        obj = response.json()
        return obj["id"]
    except Exception as e:
        logger.info(e)
        # Try org
        response = requests.get(
            f"{GITHUB_API_BASE_URL}/orgs/{username}/installation",
            headers={
                "Accept": "application/vnd.github+json",
                "Authorization": "Bearer " + jwt,
                "X-GitHub-Api-Version": "2022-11-28",
            },
        )
        try:
            obj = response.json()
            return obj["id"]
        except Exception as e:
            logger.error(e)
            logger.error(response.text)
        raise Exception("Could not get installation, probably not installed")


# for check if a file exists within a github repo (calls the actual github api)
def file_exists_in_repo(repo: Repository, filepath: str):
    try:
        # Attempt to get the contents of the file
        repo.get_contents(filepath)
        return True  # If no exception, the file exists
    except GithubException:
        return False  # File does not exist


def validate_and_sanitize_multi_file_changes(
    repo: Repository, file_changes: dict[str, str], fcrs: list[FileChangeRequest]
):
    sanitized_file_changes = {}
    all_file_names = list(file_changes.keys())
    all_fcr_file_names = set(os.path.normpath(fcr.filename) for fcr in fcrs)
    file_removed = False
    # validate each file change
    for file_name in all_file_names:
        # file_name must either appear in the repo or in a fcr
        if os.path.normpath(file_name) in all_fcr_file_names or file_exists_in_repo(repo, os.path.normpath(file_name)):
            sanitized_file_changes[file_name] = copy.deepcopy(file_changes[file_name])
        else:
            file_removed = True
    return sanitized_file_changes, file_removed


# commits multiple files in a single commit, returns the commit object
def commit_multi_file_changes(
    cloned_repo: "ClonedRepo",
    file_changes: dict[str, str],
    commit_message: str,
    branch: str,
    renames_dict: dict[str, str] = {},
):
    assert file_changes or renames_dict
    repo: Repository = cloned_repo.github_repo
    if renames_dict:
        blobs_to_commit = []
        # make a separate commit with just the renames
        for old_name, new_name in renames_dict.items():
            file_contents = cloned_repo.get_file_contents(new_name)
            blob = repo.create_git_blob(file_contents, "utf-8")
            blobs_to_commit.append(
                InputGitTreeElement(
                    path=os.path.normpath(old_name).lstrip("/"),
                    mode="100644",
                    type="blob",
                    sha=None,
                )
            )
            blobs_to_commit.append(
                InputGitTreeElement(
                    path=os.path.normpath(new_name).lstrip("/"),
                    mode="100644",
                    type="blob",
                    sha=blob.sha,
                )
            )
        head_sha = repo.get_branch(branch).commit.sha
        base_tree = repo.get_git_tree(sha=head_sha)
        # create new git tree
        new_tree = repo.create_git_tree(blobs_to_commit, base_tree=base_tree)
        # commit the changes
        parent = repo.get_git_commit(sha=head_sha)
        commit_message = "Renamed to " + ", ".join(renames_dict.values())
        commit_message = commit_message[:69] + "..." if len(commit_message) > 70 else commit_message
        commit = repo.create_git_commit(
            commit_message,
            new_tree,
            [parent],
        )
        # update ref of branch
        ref = f"heads/{branch}"
        repo.get_git_ref(ref).edit(sha=commit.sha)
        if not file_changes:
            return commit
    blobs_to_commit = []
    # convert to blob
    for path, content in file_changes.items():
        blob = repo.create_git_blob(content, "utf-8")
        blobs_to_commit.append(
            InputGitTreeElement(
                path=os.path.normpath(path).lstrip("/"),
                mode="100644",
                type="blob",
                sha=blob.sha,
            )
        )
    head_sha = repo.get_branch(branch).commit.sha
    base_tree = repo.get_git_tree(sha=head_sha)
    # create new git tree
    new_tree = repo.create_git_tree(blobs_to_commit, base_tree=base_tree)
    # commit the changes
    parent = repo.get_git_commit(sha=head_sha)
    commit = repo.create_git_commit(
        commit_message,
        new_tree,
        [parent],
    )
    # update ref of branch
    ref = f"heads/{branch}"
    repo.get_git_ref(ref).edit(sha=commit.sha)
    return commit


def clean_branch_name(branch: str) -> str:
    branch = re.sub(r"[^a-zA-Z0-9_\-/]", "-", branch)
    branch = re.sub(r"-+", "-", branch)
    branch = branch.strip("-")

    return branch


def generate_new_branch_name(repo: Repository, branch: str, base_branch: str, MAX_RETRIES: int = 10):
    """
    Try placing hashes after to get a unique branch name.
    """
    base_branch_obj = repo.get_branch(base_branch)
    for _ in range(MAX_RETRIES):
        with suppress_with_warning(GithubException):
            _hash = get_hash()[:5]
            new_branch_name = f"{branch}_{_hash}"
            logger.warning(f"Retrying {new_branch_name}...")
            repo.create_git_ref(f"refs/heads/{new_branch_name}", base_branch_obj.commit.sha)
            return repo.get_branch(new_branch_name)
    raise Exception(f"Could not create branch {branch} on {repo.full_name} after {MAX_RETRIES} retries")


def create_and_get_branch(repo: Repository, branch: str, base_branch: str = ""):
    # Creates a new branch, changes name if it already exists
    branch = clean_branch_name(branch)
    base_branch = base_branch or repo.default_branch
    try:
        repo.get_branch(branch)
    except GithubException:
        # Branch doesn't exist, expected
        base_branch_obj = repo.get_branch(base_branch)
        repo.create_git_ref(f"refs/heads/{branch}", base_branch_obj.commit.sha)
        return repo.get_branch(branch)
    else:
        # Branch exists, create a new name
        return generate_new_branch_name(repo, branch, base_branch)


REPO_CACHE_BASE_DIR = os.path.join(CACHE_DIRECTORY, "repos")


@contextmanager
def temporary_cloned_repo_copy(original_repo: ClonedRepo):
    with tempfile.TemporaryDirectory() as temp_dir:
        logger.debug(f"Copying {original_repo.repo_dir} to {temp_dir}")
        try:
            copy_with_shutil_backoff(original_repo.repo_dir, temp_dir)
            base_name = os.path.basename(original_repo.repo_dir)
            new_repo_dir = os.path.join(temp_dir, base_name)
            temp_repo = MockClonedRepo(
                _repo_dir=new_repo_dir,
                repo_full_name=original_repo.repo_full_name,
                installation_id=original_repo.installation_id,
                branch=original_repo.branch,
                token=original_repo.token,
            )
            temp_repo.git_repo.git.remote("set-url", "origin", original_repo.clone_url, env=git_env)
            yield temp_repo
        except subprocess.CalledProcessError as e:
            logger.info(f"Error copying repo: {e.stderr}")
            raise
        except Exception as e:
            logger.info(f"Unexpected error: {str(e)}")
            raise


def get_github_repo(repo_full_name: str):
    username = repo_full_name.split("/")[0]
    return CustomGithub(installation_id=get_installation_id(username=username), base_url=GITHUB_API_BASE_URL).get_repo(repo_full_name)


def has_repo_been_cloned_before(repo_full_name: str, branch: str | None = None):
    github_repo = get_github_repo(repo_full_name=repo_full_name)
    branch = branch or SweepConfig.get_branch(github_repo)
    return os.path.exists(os.path.join(REPO_CACHE_BASE_DIR, repo_full_name, "base", parse_collection_name(branch)))


def copy_with_shutil_backoff(cached_dir: str, repo_dir: str):
    """Raises the error from subprocess.run if it fails"""
    try:
        subprocess.run(
            # -R is recursive, -v is verbose, -P is preserve symlinks (no dereferencing)
            ["cp", "-Rv", "-P", cached_dir, repo_dir],
            check=True,
            capture_output=True,
            text=True,
        )
    except Exception as e1:
        # backoff to shutil if cp fails
        try:
            subprocess.run(["rm", "-rf", repo_dir], check=True)
            logger.warning(f"[YELLOW] Error copying repo: {e1}, falling back to shutil")
            shutil.copytree(src=cached_dir, dst=repo_dir, symlinks=True)
        except Exception as e2:
            logger.error(f"Error copying repo: {e1}, {e2}")
            raise e1


@dataclass
class ClonedRepo:
    repo_full_name: str
    branch: str | None = None
    installation_id: str | None = None
    token: str | None = None
    github_repo: Repository | None = None
    git_repo: git.Repo | None = None

    class Config:
        arbitrary_types_allowed = True

    @cached_property
    def cached_dir(self):
        return os.path.join(
            REPO_CACHE_BASE_DIR,
            self.repo_full_name,
            "base",
            self.default_branch,
        )

    @cached_property
    def repo_dir(self):  # pylint: disable=method-hidden
        commit_sha = (
            self.github_repo.get_branch(self.branch).commit.sha
            or self.github_repo.get_commit(self.branch).sha
            or "latest"
        )
        return os.path.join(
            REPO_CACHE_BASE_DIR,
            self.repo_full_name,
            commit_sha,
        )

    @property
    def repo_name(self):
        # just get the repo name from the repo_full_name without org_name
        return self.repo_full_name.split("/")[1]

    @property
    def clone_url(self):
        return f"https://x-access-token:{self.token}@{GITHUB_BASE_URL}/{self.repo_full_name}.git"

    @cached_property
    def current_commit_sha(self):
        return self.git_repo.git.rev_parse("HEAD")

    @cached_property
    def cache_key(self):
        return f"{self.repo_full_name}:{self.current_commit_sha}"

    @property
    def default_branch(self):
        return self.github_repo.default_branch

    def clone(self):
        if is_valid_git_repo(self.cached_dir):
            git_repo = git.Repo(self.cached_dir)
        else:
            subprocess.run(["rm", "-rf", self.cached_dir])
            git_repo = git.Repo.clone_from(
                self.clone_url,
                self.cached_dir,
                env=git_env,
                multi_options=["--recurse-submodules"],
            )

        # Check whether cached_dir needs to be updated
        latest_commit = self.github_repo.get_commit(self.default_branch).sha
        cached_dir_commit = git_repo.git.rev_parse("HEAD")
        should_pull_latest_cached_dir = not (latest_commit == cached_dir_commit)
        if should_pull_latest_cached_dir:
            logger.info("Cached directory is outdated, pulling cached_dir...")
            try:
                git_repo.git.remote("set-url", "origin", self.clone_url, env=git_env)
                git_repo.git.clean("-fd")
                git_repo.git.checkout(self.default_branch, env=git_env)
                git_repo.git.reset("--hard", f"origin/{self.default_branch}", env=git_env)
                git_repo.git.pull(env=git_env)
            except Exception as e:
                logger.error(f"Error pulling cached_dir: {e}")
                raise Exception("Error pulling cached_dir")

        if is_valid_git_repo(self.repo_dir):
            git_repo = git.Repo(self.repo_dir)
        else:
            logger.warning(f"Invalid git repository: {self.repo_dir}")
            self.delete_and_copy_latest_cached_dir_to_repo_dir()
            git_repo = git.Repo(self.repo_dir)
        
        try:
            git_repo.git.remote("set-url", "origin", self.clone_url, env=git_env)
        except Exception as e:
            logger.error(f"Error setting remote URL: {e}")
        return git_repo

    def __post_init__(self):
        subprocess.run(["git", "config", "--global", "http.postBuffer", "524288000"])
        if self.installation_id:
            self.token = self.token or get_token(self.installation_id)
        # Try to get the github repo from cache
        self.github_repo = self.github_repo or get_github_repo(repo_full_name=self.repo_full_name)
        self.branch = self.branch or SweepConfig.get_branch(self.github_repo)
        self.git_repo: git.Repo = self.clone()
        all_repos.append(self)
        if self.branch != self.default_branch:
            self.safely_checkout_to_non_default_branch(self.github_repo.get_branch(self.branch).commit.sha)

    def delete_and_copy_latest_cached_dir_to_repo_dir(self):
        if os.path.exists(self.repo_dir):
            logger.info(f"Repo directory {self.repo_dir} exists but is empty, deleting...")
            subprocess.run(["rm", "-rf", self.repo_dir])
        logger.info("Copying repo...")
        logger.debug(f"Copying {self.cached_dir} to {self.repo_dir}")
        os.makedirs(os.path.dirname(self.repo_dir), exist_ok=True)
        try:
            copy_with_shutil_backoff(self.cached_dir, self.repo_dir)
        except Exception as e:
            # we need this check since sometimes gitignored files still "exist" via symlinks
            if "No such file or directory" in str(e):
                pass
            else:
                logger.info(f"Error copying repo {str(e)}")
                raise e
        assert os.path.exists(self.repo_dir) and len(os.listdir(self.repo_dir)) > 0
        logger.info("Done copying")

    def handle_checkout_failures(self):
        untracked_files = self.git_repo.untracked_files
        if untracked_files:
            logger.info(f"Untracked files found: {', '.join(untracked_files)}")
            for file in untracked_files:
                file_path = os.path.join(self.git_repo.working_dir, file)
                if os.path.isfile(file_path):
                    logger.info(f"Removing untracked file: {file}")
                    os.remove(file_path)
                elif os.path.isdir(file_path):
                    logger.info(f"Removing untracked directory: {file}")
                    os.removedirs(file_path)
        else:
            logger.info("No untracked files found")

        # if the index.lock file exists, delete it
        index_lock_file = os.path.join(self.git_repo.working_dir, "index.lock")
        if os.path.exists(index_lock_file):
            logger.info(f"Removing index lock file: {index_lock_file}")
            os.remove(index_lock_file)

        logger.info("Cleaning untracked files")
        self.git_repo.git.clean("-fd")

    @classmethod
    def from_repo_full_name(cls, repo_full_name: str, branch: str | None = None):
        g = CustomGithub.from_repo_full_name(repo_full_name)
        repo = g.get_repo(repo_full_name)
        branch = branch or repo.default_branch
        return cls(
            repo_full_name,
            branch=branch,
            installation_id=g.installation_id,
            token=g.token,
            github_repo=repo,
        )

    def safely_checkout_to_non_default_branch(self, branch_commit_sha: str):
        branch_commit_repo_dir = os.path.join(REPO_CACHE_BASE_DIR, self.repo_full_name, branch_commit_sha)
        self.repo_dir = branch_commit_repo_dir
        # Ensure the git repo is valid, copy it over if not
        if not is_valid_git_repo(self.repo_dir):
            try:
                # copy the cached_dir to the repo_dir
                self.delete_and_copy_latest_cached_dir_to_repo_dir()
            except subprocess.CalledProcessError as e:
                error_message = f"Command '{e.cmd}' failed with exit status {e.returncode}.\n"
                error_message += f"STDOUT: {e.stdout}\n"
                error_message += f"STDERR: {e.stderr}"
                logger.error(f"Error copying repo: {error_message}")
                raise e
        # This should always be a valid git_repo now
        self.git_repo = git.Repo(self.repo_dir)
        # Checkout to the branch_commit_sha and setup the origin
        try:
            self.git_repo.git.remote("set-url", "origin", self.clone_url, env=git_env)
            self.git_repo.git.fetch("origin", branch_commit_sha, env=git_env)
            self.git_repo.git.checkout(branch_commit_sha, env=git_env)
        except Exception as e:
            logger.error(f"Error checking out branch {branch_commit_sha}: {e}")
            raise
        return

    def delete_repo_on_disk(self):
        # you can't use any packages here since this is called during __del__, so os becomes a NoneType and os.path is a NPE
        try:
            if os.path.exists(self.repo_dir) and not isinstance(self.repo_dir, tempfile.TemporaryDirectory):
                logger.info(f"Deleting repo {self.repo_dir}")
                subprocess.Popen(
                    ["rm", "-rf", self.repo_dir],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    stdin=subprocess.DEVNULL,
                    preexec_fn=os.setsid,
                )
                return True
            else:
                logger.info(f"Repo {self.repo_dir} does not exist or is a TemporaryDirectory.")
                return False
        except Exception as e:
            # I use print because on shutdown loguru might have been set to none already
            print(f"Failed to delete {self.repo_dir}: {str(e)}")
            return False

    @contextmanager
    def cleanup_context(self):
        try:
            yield self
        finally:
            self.delete_repo_on_disk()

    def get_file_list(self):
        root_directory = self.repo_dir
        files: list[str] = []
        sweep_config: SweepConfig = SweepConfig()

        def dfs_helper(directory):
            nonlocal files
            for item in os.listdir(directory):
                if item == ".git":
                    continue
                if item in sweep_config.exclude_dirs:  # this saves a lot of time
                    continue
                item_path = os.path.join(directory, item)
                if os.path.isfile(item_path):
                    # make sure the item_path is not in one of the banned directories
                    if not sweep_config.is_file_excluded(item_path):
                        files.append(item_path)  # Add the file to the list
                elif os.path.isdir(item_path):
                    dfs_helper(item_path)  # Recursive call to explore subdirectory

        dfs_helper(root_directory)
        files = [file[len(root_directory) + 1 :] for file in files]
        return files

    def get_directory_list(self):
        root_directory = self.repo_dir
        files = []
        sweep_config: SweepConfig = SweepConfig()

        def dfs_helper(directory):
            nonlocal files
            for item in os.listdir(directory):
                if item == ".git":
                    continue
                if item in sweep_config.exclude_dirs:  # this saves a lot of time
                    continue
                item_path = os.path.join(directory, item)
                if os.path.isdir(item_path):
                    files.append(item_path)  # Add the file to the list
                    dfs_helper(item_path)  # Recursive call to explore subdirectory

        dfs_helper(root_directory)
        files = [file[len(root_directory) + 1 :] for file in files]
        return files

    def get_immediate_subdirs(self):
        # fast way to get all immediate subdirs without a dfs
        return [name for name in os.listdir(self.repo_dir) if os.path.isdir(os.path.join(self.repo_dir, name))]

    def get_file_contents(self, file_path, ref=None):
        local_path = os.path.join(self.repo_dir, file_path.lstrip("/"))
        if os.path.exists(local_path) and os.path.isfile(local_path):
            with open(local_path, "r", encoding="utf-8", errors="replace") as f:
                contents = f.read()
            return contents
        else:
            raise FileNotFoundError(f"{local_path} does not exist.")

    def get_num_files_from_repo(self):
        # subprocess.run(["git", "config", "--global", "http.postBuffer", "524288000"])
        self.git_repo.git.checkout(self.default_branch)
        file_list = self.get_file_list()
        return len(file_list)

    def get_commit_history(self, username: str = "", limit: int = 200, time_limited: bool = True):
        commit_history = []
        try:
            if username != "":
                commit_list = list(self.git_repo.iter_commits(author=username))
            else:
                commit_list = list(self.git_repo.iter_commits())
            line_count = 0
            cut_off_date = datetime.now() - timedelta(days=7)
            for commit in commit_list:
                # must be within a week
                if time_limited and commit.authored_datetime.replace(tzinfo=None) <= cut_off_date.replace(tzinfo=None):
                    logger.info("Exceeded cut off date, stopping...")
                    break
                repo = get_github_client(self.installation_id)[1].get_repo(self.repo_full_name)
                branch = SweepConfig.get_branch(repo)
                if branch not in self.git_repo.git.branch():
                    branch = f"origin/{branch}"
                diff = self.git_repo.git.diff(commit, branch, unified=1)
                lines = diff.count("\n")
                # total diff lines must not exceed 200
                if lines + line_count > limit:
                    logger.info(f"Exceeded {limit} lines of diff, stopping...")
                    break
                commit_history.append(
                    f"<commit>\nAuthor: {commit.author.name}\nMessage: {commit.message}\n{diff}\n</commit>"
                )
                line_count += lines
        except Exception:
            logger.error(f"An error occurred: {traceback.print_exc()}")
        return commit_history

    def get_similar_directories(self, file_path: str, limit: int = 5):
        from rapidfuzz.fuzz import QRatio

        # Fuzzy search over file names
        # file_name = os.path.basename(file_path)
        all_file_paths = self.get_directory_list()

        # get top limit similar directories
        sorted_file_paths = sorted(
            all_file_paths,
            key=lambda file_name_: QRatio(file_name_, file_path),
            reverse=True,
        )

        filtered_file_paths = list(filter(lambda file_name_: QRatio(file_name_, file_path) > 50, sorted_file_paths))

        return filtered_file_paths[:limit]

    def get_similar_file_paths(self, file_path: str, limit: int = 10):
        from rapidfuzz.fuzz import ratio

        # Fuzzy search over file names
        file_name = os.path.basename(file_path)
        all_file_paths = self.get_file_list()
        # filter for matching extensions if both have extensions
        if "." in file_name:
            all_file_paths = [
                file for file in all_file_paths if "." in file and file.split(".")[-1] == file_name.split(".")[-1]
            ]
        files_with_matching_name = []
        files_without_matching_name = []
        for file_path in all_file_paths:
            if file_name in file_path:
                files_with_matching_name.append(file_path)
            else:
                files_without_matching_name.append(file_path)
        file_path_to_ratio = {file: ratio(file_name, file) for file in all_file_paths}
        files_with_matching_name = sorted(
            files_with_matching_name,
            key=lambda file_path: file_path_to_ratio[file_path],
            reverse=True,
        )
        files_without_matching_name = sorted(
            files_without_matching_name,
            key=lambda file_path: file_path_to_ratio[file_path],
            reverse=True,
        )
        # this allows 'config.py' to return 'sweepai/config/server.py', 'sweepai/config/client.py', 'sweepai/config/__init__.py' and no more
        filtered_files_without_matching_name = list(
            filter(
                lambda file_path: file_path_to_ratio[file_path] > 50,
                files_without_matching_name,
            )
        )
        all_files = files_with_matching_name + filtered_files_without_matching_name
        return all_files[:limit]


# updates a file with new_contents, returns True if successful
def update_file(root_dir: str, file_path: str, new_contents: str):
    local_path = os.path.join(root_dir, file_path)
    try:
        with open(local_path, "w") as f:
            f.write(new_contents)
        return True
    except Exception as e:
        logger.error(f"Failed to update file: {e}")
        return False

def is_valid_git_repo(repo_path: str) -> bool:
    """
    Checks if a directory is a valid git repository by verifying:
    1. .git directory exists and has required files
    2. git commands can be executed
    3. repository has a valid HEAD reference
    4. repository has valid objects and refs
    """
    try:
        # Check if directory exists
        if not os.path.exists(repo_path):
            return False
            
        # Check for .git directory and critical files
        git_dir = os.path.join(repo_path, ".git")
        required_git_files = [
            os.path.join(git_dir, "HEAD"),  # HEAD reference file
            os.path.join(git_dir, "config"),  # Git config
            os.path.join(git_dir, "refs"),  # refs directory
            os.path.join(git_dir, "objects"),  # objects directory
        ]
        
        if not all(os.path.exists(f) for f in required_git_files):
            return False
            
        # Try to initialize repo object
        repo = git.Repo(repo_path)
        
        # Check if HEAD is valid
        if repo.head.is_valid():
            try:
                # Try to get current branch name
                _ = repo.active_branch.name
            except TypeError:
                # HEAD exists but might be detached
                pass
        
        return True
        
    except (git.InvalidGitRepositoryError, git.NoSuchPathError) as e:
        logger.warning(f"Invalid git repository: {str(e)}")
        return False
    except Exception as e:
        logger.error(f"Unexpected error while validating git repo: {str(e)}")
        return False


@dataclass
class MockClonedRepo(ClonedRepo):
    _repo_dir: str = ""
    git_repo: git.Repo | None = None

    def __init__(
        self,
        _repo_dir: str,
        repo_full_name: str,
        installation_id: str = "",
        branch: str | None = None,
        token: str | None = None,
        repo: Repository | None = None,
    ):
        self._repo_dir = _repo_dir
        self.repo_full_name = repo_full_name
        self.installation_id = installation_id
        self.branch = branch
        self.token = token
        if repo:
            self.github_repo = repo
        else:
            try:
                org_name, repo_name = self.repo_full_name.split("/")
                installation = github_integration.get_repo_installation(org_name, repo_name)
                self.github_repo = installation.get_github_for_installation().get_repo(self.repo_full_name)
            except Exception as e:
                logger.warning(f"[YELLOW] Failed to get github repo for {self.repo_full_name}: {str(e)}")
                self.github_repo = None
        self.git_repo = git.Repo(self.repo_dir)

    @classmethod
    def from_dir(cls, repo_dir: str, **kwargs):
        return cls(_repo_dir=repo_dir, **kwargs)

    @property
    def cached_dir(self):
        return self._repo_dir

    @property
    def repo_dir(self):
        return self._repo_dir

    def clone(self):
        return git.Repo(self.repo_dir)

    def __post_init__(self):
        return self

    def __del__(self):
        return True


def parse_collection_name(name: str) -> str:
    # Replace any non-alphanumeric characters with hyphens
    name = re.sub(r"[^\w-]", "--", name)
    # Ensure the name is between 3 and 63 characters and starts/ends with alphanumeric
    name = re.sub(r"^(-*\w{0,61}\w)-*$", r"\1", name[:63].ljust(3, "x"))
    return name


# set whether or not a pr is a draft, there is no way to do this using pygithub
def convert_pr_draft_field(pr: PullRequest, is_draft: bool = False, installation_id: int = 0) -> bool:
    token = get_token(installation_id)
    pr_id = pr.raw_data["node_id"]
    # GraphQL mutation for marking a PR as ready for review
    mutation = """
    mutation MarkPRReady {
    markPullRequestReadyForReview(input: {pullRequestId: {pull_request_id}}) {
    pullRequest {
    id
    }
    }
    }
    """.replace(
        "{pull_request_id}", '"' + pr_id + '"'
    )

    # GraphQL API URL
    url = f"{GITHUB_API_BASE_URL}/graphql"

    # Headers
    headers = {
        "Accept": "application/vnd.github+json",
        "Authorization": f"Bearer {token}",
        "X-GitHub-Api-Version": "2022-11-28",
    }

    # Prepare the JSON payload
    json_data = {
        "query": mutation,
    }

    # Make the POST request
    response = requests.post(url, headers=headers, data=json.dumps(json_data))
    if response.status_code != 200:
        logger.error(f"Failed to convert PR to {'draft' if is_draft else 'open'}")
        return False
    return True


### REVIEW RELATED


# get the review threads obejct for a pr, required to tell if a comment is resolved or not
def get_review_threads(repo_full_name: str, pr_number: int, installation_id: int):
    token = get_token(installation_id)
    query = """
query GetReviewThreads($owner: String!, $name: String!, $prNumber: Int!) {
  repository(owner: $owner, name: $name) {
    pullRequest(number: $prNumber) {
      reviewThreads(first: 100) {
        nodes {
          id
          isResolved
          isOutdated
          path
          line
          comments(first: 100) {
            nodes {
              id
              body
              author {
                login
              }
              createdAt
            }
          }
        }
      }
    }
  }
}
    """

    # GraphQL API URL
    url = f"{GITHUB_API_BASE_URL}/graphql"

    # Headers
    headers = {
        "Accept": "application/vnd.github+json",
        "Authorization": f"Bearer {token}",
        "X-GitHub-Api-Version": "2022-11-28",
    }
    owner, name = repo_full_name.split("/")
    # Prepare the JSON payload
    variables = {"owner": owner, "name": name, "prNumber": pr_number}
    json_data = {"query": query, "variables": variables}

    # Make the POST request
    response = requests.post(url, headers=headers, data=json.dumps(json_data))
    if response.status_code != 200:
        return {}
    review_threads_json = response.json()["data"]["repository"]["pullRequest"]["reviewThreads"]["nodes"]
    return review_threads_json


# refresh user token, github client and repo object
def refresh_token(repo_full_name: str, installation_id: int):
    user_token, g = get_github_client(installation_id)
    repo = g.get_repo(repo_full_name)
    return user_token, g, repo


if __name__ == "__main__":
    try:
        organization_name = "sweepai"
        sweep_config = SweepConfig()
        installation_id = get_installation_id(organization_name)
        user_token, g = get_github_client(installation_id)
        cloned_repo = ClonedRepo("sweepai/sweep", installation_id, "main")
        commit_history = cloned_repo.get_commit_history()
        similar_file_paths = cloned_repo.get_similar_file_paths("config.py")
        # ensure no similar file_paths are sweep excluded
        assert not any([file for file in similar_file_paths if sweep_config.is_file_excluded(file)])
        print(f"similar_file_paths: {similar_file_paths}")
        str1 = "a\nline1\nline2\nline3\nline4\nline5\nline6\ntest\n"
        str2 = "a\nline1\nlineTwo\nline3\nline4\nline5\nlineSix\ntset\n"
        mocked_repo = MockClonedRepo.from_dir(
            cloned_repo.repo_dir,
            repo_full_name="sweepai/sweep",
        )
        print(f"mocked repo: {mocked_repo}")
    except Exception as e:
        logger.error(f"github_utils.py failed to run successfully with error: {e}")
